"""
Color Palette

A color palette is defined by its swatches. A swatch is an appropriate
sampling of colors in the palette. To maximise the difference between
colors, it is often better to choose the shortest swatch possible. This
means that the palette space is sparsely sampled.

All swatches of a palette have different lengths, are listed shortest to
longest and each swatch is longer than the previous by one color.
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from .._colormaps import PaletteInterpolatedMap
from .._colormaps._colormap import ColorMapKind

if TYPE_CHECKING:
    from mizani.typing import (
        RGB256Color,
        RGB256Swatch,
        RGB256Swatches,
        RGBHexColor,
        RGBHexSwatch,
    )

PaletteKind = ColorMapKind


@dataclass
class palette:
    #: Name of palette
    name: str

    #: Number of colors in the shortest swatch
    min_colors: int

    #: Number of colors in the longest swatch
    max_colors: int

    #: Discrete samplings of the palette space
    swatches: RGB256Swatches

    #: Kind of palette
    kind: PaletteKind

    def get_swatch(self, num_colors: int) -> RGB256Swatch:
        """
        Get a swatch with given number of colors
        """
        index = num_colors - self.min_colors
        return self.swatches[index]

    def get_hex_swatch(self, num_colors: int) -> RGBHexSwatch:
        """
        Get a swatch with given number of colors in hex
        """
        swatch = self.get_swatch(num_colors)
        return RGB256Swatch_to_RGBHexSwatch(swatch)

    @cached_property
    def colormap(self) -> PaletteInterpolatedMap:
        """
        Return a colormap representation of the palette
        """
        return PaletteInterpolatedMap(self)


def HX(n: int) -> str:
    """
    Conver 8-Bit int to two character HEX (uppercase)

    Should be in the range [0, 255]
    """
    return f"{hex(n)[2:]:>02}".upper()


def RGB256Color_to_RGBHexColor(color: RGB256Color) -> RGBHexColor:
    """
    Covert 256Color to HexColor
    """
    return "#" + "".join(HX(i) for i in color)


def RGB256Swatch_to_RGBHexSwatch(swatch: RGB256Swatch) -> RGBHexSwatch:
    """
    Covert 256Swatch to HexSwatch
    """
    return [RGB256Color_to_RGBHexColor(color) for color in swatch]



"""
Scales have guides and these are what help users make sense of
the data mapped onto the scale. Common examples of guides include
the x-axis, the y-axis, the keyed legend and a colorbar legend.
The guides have demarcations(breaks), some of which must be labelled.

The `label_*` functions below create functions that convert data
values as understood by a specific scale and return string
representations of those values. Manipulating the string
representation of a value helps improve readability of the guide.
"""

from __future__ import annotations

import re
from bisect import bisect_right
from dataclasses import dataclass
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

import numpy as np

from .breaks import timedelta_helper
from .utils import (
    match,
    precision,
    round_any,
    same_log10_order_of_magnitude,
)

if TYPE_CHECKING:
    from datetime import datetime, tzinfo
    from typing import Literal, Optional, Sequence

    from mizani.typing import (
        BytesSymbol,
        DurationUnit,
        FloatArrayLike,
        NDArrayTimedelta,
        TupleInt2,
    )

__all__ = [
    "label_comma",
    "label_custom",
    "label_currency",
    "label_dollar",
    "label_percent",
    "label_scientific",
    "label_date",
    "label_number",
    "label_log",
    "label_timedelta",
    "label_pvalue",
    "label_ordinal",
    "label_bytes",
]

UTC = ZoneInfo("UTC")


@dataclass
class label_number:
    """
    Labelling numbers

    Parameters
    ----------
    precision : int
        Number of digits after the decimal point.
    suffix : str
        What to put after the value.
    big_mark : str
        The thousands separator. This is usually
        a comma or a dot.
    decimal_mark : str
        What to use to separate the decimals digits.

    Examples
    --------
    >>> label_number()([.654, .8963, .1])
    ['0.65', '0.90', '0.10']
    >>> label_number(accuracy=0.0001)([.654, .8963, .1])
    ['0.6540', '0.8963', '0.1000']
    >>> label_number(precision=4)([.654, .8963, .1])
    ['0.6540', '0.8963', '0.1000']
    >>> label_number(prefix="$")([5, 24, -42])
    ['$5', '$24', '-$42']
    >>> label_number(suffix="s")([5, 24, -42])
    ['5s', '24s', '-42s']
    >>> label_number(big_mark="_")([1e3, 1e4, 1e5, 1e6])
    ['1_000', '10_000', '100_000', '1_000_000']
    >>> label_number(width=3)([1, 10, 100, 1000])
    ['  1', ' 10', '100', '1000']
    >>> label_number(align="^", width=5)([1, 10, 100, 1000])
    ['  1  ', ' 10  ', ' 100 ', '1000 ']
    >>> label_number(style_positive=" ")([5, 24, -42])
    [' 5', ' 24', '-42']
    >>> label_number(style_positive="+")([5, 24, -42])
    ['+5', '+24', '-42']
    >>> label_number(prefix="$", style_negative="braces")([5, 24, -42])
    ['$5', '$24', '($42)']
    """

    accuracy: Optional[float] = None
    precision: Optional[int] = None
    scale: float = 1
    prefix: str = ""
    suffix: str = ""
    big_mark: str = ""
    decimal_mark: str = "."
    fill: str = ""
    style_negative: Literal["-", "hyphen", "parens"] = "-"
    style_positive: Literal["", "+", " "] = ""
    align: Literal["<", ">", "=", "^"] = ">"
    width: Optional[int] = None

    def __post_init__(self):
        if self.precision is not None:
            if self.accuracy is not None:
                raise ValueError("Specify only one of precision or accuracy")
            self.accuracy = 10**-self.precision

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        # Construct formatting according to
        # https://docs.python.org/3/library/string.html#format-string-syntax
        # Specfically using the Format Specification Mini-Language

        # python format only accepts ",", "_" to separate the thousands
        # if we have a non-standard value, we use "," & replace it after
        valid_big_mark = self.big_mark in ("", ",", "_")
        sep = self.big_mark if valid_big_mark else ","

        fmt = (
            f"{self.prefix}" f"{{num:{sep}.{{precision}}f}}" f"{self.suffix}"
        ).format

        x = np.asarray(x)
        x_scaled = x * self.scale

        if self.accuracy is None:
            accuracy = precision(x_scaled)
        else:
            accuracy = self.accuracy

        x = round_any(x, accuracy / self.scale)
        digits = -np.floor(np.log10(accuracy)).astype(int)
        digits = np.minimum(np.maximum(digits, 0), 20)

        res = [fmt(num=abs(n), precision=digits) for n in x_scaled]
        if not valid_big_mark:
            res = [s.replace(",", self.big_mark) for s in res]

        if self.decimal_mark != ".":
            res = [s.replace(".", self.decimal_mark) for s in res]

        pos_fmt = f"{self.style_positive}{{s}}".format

        if self.style_negative == "-":
            neg_fmt = "-{s}".format
        elif self.style_negative == "hyphen":
            neg_fmt = "\u2212{s}".format
        else:
            neg_fmt = "({s})".format

        res = [
            neg_fmt(s=s) if num < 0 else pos_fmt(s=s) for num, s in zip(x, res)
        ]

        if self.width is not None:
            fmt = f"{{s:{self.fill}{self.align}{self.width}}}".format
            res = [fmt(s=s) for s in res]

        return res


@dataclass
class label_custom:
    """
    Creating a custom labelling function

    Parameters
    ----------
    fmt : str, optional
        Format string. Default is the generic new style
        format braces, ``{}``.
    style : 'new' | 'old'
        Whether to use new style or old style formatting.
        New style uses the :meth:`str.format` while old
        style uses ``%``. The format string must be written
        accordingly.

    Examples
    --------
    >>> label = label_custom('{:.2f} USD')
    >>> label([3.987, 2, 42.42])
    ['3.99 USD', '2.00 USD', '42.42 USD']
    """

    fmt: str = "{}"
    style: Literal["old", "new"] = "new"

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        if self.style == "new":
            return [self.fmt.format(val) for val in x]
        elif self.style == "old":
            return [self.fmt % val for val in x]
        else:
            raise ValueError("style should be either 'new' or 'old'")


# formatting functions
@dataclass
class label_currency(label_number):
    """
    Labelling currencies

    Parameters
    ----------
    prefix : str
        What to put before the value.

    Examples
    --------
    >>> x = [1.232, 99.2334, 4.6, 9, 4500]
    >>> label_currency()(x)
    ['$1.23', '$99.23', '$4.60', '$9.00', '$4500.00']
    >>> label_currency(prefix='C$', precision=0, big_mark=',')(x)
    ['C$1', 'C$99', 'C$5', 'C$9', 'C$4,500']
    """

    prefix: str = "$"

    def __post_init__(self):
        if self.precision is None and self.accuracy is None:
            self.precision = 2
        super().__post_init__()


label_dollar = label_currency
dollar = label_dollar()


@dataclass
class label_comma(label_currency):
    """
    Labels of numbers with commas as separators

    Parameters
    ----------
    precision : int
        Number of digits after the decimal point.

    Examples
    --------
    >>> label_comma()([1000, 2, 33000, 400])
    ['1,000', '2', '33,000', '400']
    """

    prefix: str = ""
    precision: int = 0
    big_mark: str = ","


@dataclass
class label_percent(label_number):
    """
    Labelling percentages

    Multiply by one hundred and display percent sign

    Examples
    --------
    >>> label = label_percent()
    >>> label([.45, 9.515, .01])
    ['45%', '952%', '1%']
    >>> label([.654, .8963, .1])
    ['65%', '90%', '10%']
    """

    scale: float = 100
    suffix: str = "%"


percent = label_percent()


@dataclass
class label_scientific:
    """
    Scientific number labels

    Parameters
    ----------
    digits : int
        Significant digits.

    Examples
    --------
    >>> x = [.12, .23, .34, 45]
    >>> label_scientific()(x)
    ['1.2e-01', '2.3e-01', '3.4e-01', '4.5e+01']

    Notes
    -----
    Be careful when using many digits (15+ on a 64
    bit computer). Consider of the `machine epsilon`_.

    .. _machine epsilon: https://en.wikipedia.org/wiki/Machine_epsilon
    """

    digits: int = 3

    def __post_init__(self):
        tpl = f"{{:.{self.digits}e}}"
        self._label = label_custom(tpl)
        self.trailling_zeros_pattern = re.compile(r"(0+)e")

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        if len(x) == 0:
            return []

        def count_zeros(s):
            match = self.trailling_zeros_pattern.search(s)
            if match:
                return len(match.group(1))
            else:
                return 0

        # format and then remove superfluous zeros
        labels = self._label(x)
        n = min([count_zeros(val) for val in labels])
        if n:
            labels = [val.replace("0" * n + "e", "e") for val in labels]
        return labels


scientific = label_scientific()


@dataclass
class label_log:
    """
    Log number labels

    Parameters
    ----------
    base : int
        Base of the logarithm. Default is 10.
    exponent_limits : tuple
        limits (int, int) where if the any of the powers of the
        numbers falls outside, then the labels will be in
        exponent form. This only applies for base 10.
    mathtex : bool
        If True, return the labels in mathtex format as understood
        by Matplotlib.

    Examples
    --------
    >>> label_log()([0.001, 0.1, 100])
    ['0.001', '0.1', '100']

    >>> label_log()([0.0001, 0.1, 10000])
    ['1e-4', '1e-1', '1e4']

    >>> label_log(mathtex=True)([0.0001, 0.1, 10000])
    ['$10^{-4}$', '$10^{-1}$', '$10^{4}$']
    """

    base: float = 10
    exponent_limits: TupleInt2 = (-4, 4)
    mathtex: bool = False

    def _tidyup_labels(self, labels: Sequence[str]) -> Sequence[str]:
        """
        Make all labels uniform in format

        Remove redundant zeros for labels in exponential format.

        Parameters
        ----------
        labels : list-like
            Labels to be tidied.

        Returns
        -------
        out : list-like
            Labels
        """

        def remove_zeroes(s: str) -> str:
            """
            Remove unnecessary zeros for float string s
            """
            tup = s.split("e")
            if len(tup) == 2:
                mantissa = tup[0].rstrip("0").rstrip(".")
                exponent = int(tup[1])
                s = f"{mantissa}e{exponent}" if exponent else mantissa
            return s

        def as_exp(s: str) -> str:
            """
            Float string s as in exponential format
            """
            return s if "e" in s else "{:1.0e}".format(float(s))

        def as_mathtex(s: str) -> str:
            """
            Mathtex for maplotlib
            """
            if "e" not in s:
                assert s == "1", f"Unexpected value {s = }, instead of '1'"
                return f"${self.base}^{{0}}$"

            exp = s.split("e")[1]
            return f"${self.base}^{{{exp}}}$"

        # If any are in exponential format, make all of
        # them expontential
        has_e = ["e" in x for x in labels]
        if not all(has_e) and sum(has_e):
            labels = [as_exp(x) for x in labels]

        labels = [remove_zeroes(x) for x in labels]

        has_e = ["e" in x for x in labels]
        if self.mathtex and any(has_e):
            labels = [as_mathtex(x) for x in labels]

        return labels

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        if len(x) == 0:
            return []

        # Decide on using exponents
        if self.base == 10:
            xmin = int(np.floor(np.log10(np.min(x))))
            xmax = int(np.ceil(np.log10(np.max(x))))
            emin, emax = self.exponent_limits
            all_multiples = np.all([np.log10(num).is_integer() for num in x])
            beyond_threshold = xmin <= emin or emax <= xmax
            use_exponents = (
                same_log10_order_of_magnitude(x) or all_multiples
            ) and beyond_threshold
            fmt = "{:1.0e}" if use_exponents else "{:g}"
            labels = [fmt.format(num) for num in x]
            return self._tidyup_labels(labels)
        else:

            def _exp(num, base):
                e = np.log(num) / np.log(base)
                e_round = np.round(e)
                e = int(e_round) if np.isclose(e, e_round) else np.round(e, 3)
                return e

            base_txt = f"{self.base}"
            if self.base == np.e:
                base_txt = "e"

            if self.mathtex:
                fmt_parts = (f"${base_txt}^", "{{{e}}}$")
            else:
                fmt_parts = (f"{base_txt}^", "{e}")

            fmt = "".join(fmt_parts)
            exps = [_exp(num, self.base) for num in x]
            labels = [fmt.format(e=e) for e in exps]
            return labels


@dataclass
class label_date:
    """
    Datetime labels

    Parameters
    ----------
    fmt : str
        Format string. See
        :ref:`strftime <strftime-strptime-behavior>`.
    tz : datetime.tzinfo, optional
        Time zone information. If none is specified, the
        time zone will be that of the first date. If the
        first date has no time information then a time zone
        is chosen by other means.

    Examples
    --------
    >>> from datetime import datetime
    >>> x = [datetime(x, 1, 1) for x in [2010, 2014, 2018, 2022]]
    >>> label_date()(x)
    ['2010-01-01', '2014-01-01', '2018-01-01', '2022-01-01']
    >>> label_date('%Y')(x)
    ['2010', '2014', '2018', '2022']

    Can format time

    >>> x = [datetime(2017, 12, 1, 16, 5, 7)]
    >>> label_date("%Y-%m-%d %H:%M:%S")(x)
    ['2017-12-01 16:05:07']

    Time zones are respected

    >>> UTC = ZoneInfo('UTC')
    >>> UG = ZoneInfo('Africa/Kampala')
    >>> x = [datetime(2010, 1, 1, i) for i in [8, 15]]
    >>> x_tz = [datetime(2010, 1, 1, i, tzinfo=UG) for i in [8, 15]]
    >>> label_date('%Y-%m-%d %H:%M')(x)
    ['2010-01-01 08:00', '2010-01-01 15:00']
    >>> label_date('%Y-%m-%d %H:%M')(x_tz)
    ['2010-01-01 08:00', '2010-01-01 15:00']

    Format with a specific time zone

    >>> label_date('%Y-%m-%d %H:%M', tz=UTC)(x_tz)
    ['2010-01-01 05:00', '2010-01-01 12:00']
    >>> label_date('%Y-%m-%d %H:%M', tz='EST')(x_tz)
    ['2010-01-01 00:00', '2010-01-01 07:00']
    """

    fmt: str = "%Y-%m-%d"
    tz: Optional[tzinfo] = None

    def __post_init__(self):
        if isinstance(self.tz, str):
            self.tz = ZoneInfo(self.tz)

    def __call__(self, x: Sequence[datetime]) -> Sequence[str]:
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        if self.tz is not None:
            x = [d.astimezone(self.tz) for d in x]
        return [d.strftime(self.fmt) for d in x]


@dataclass
class label_timedelta:
    """
    Timedelta labels

    Parameters
    ----------
    units : str, optional
        The units in which the breaks will be computed.
        If None, they are decided automatically. Otherwise,
        the value should be one of::

            'ns'    # nanoseconds
            'us'    # microseconds
            'ms'    # milliseconds
            's'     # seconds
            'min'   # minute
            'h'     # hour
            'day'     # day
            'week'  # week
            'month' # month
            'year'  # year

    show_units : bool
        Whether to append the units symbol to the values.
    zero_has_units : bool
        If True a value of zero
    usetex : bool
        If True, they microseconds identifier string is
        rendered with greek letter *mu*. Default is False.
    space : bool
        If True add a space between the value and the units
    use_plurals : bool
        If True, for the when the value is not 1 and the units are
        one of `week`, `month` and `year`, the plural form of the
        unit is used e.g. `2 weeks`.

    Examples
    --------
    >>> from datetime import timedelta
    >>> x = [timedelta(days=31*i) for i in range(5)]
    >>> label_timedelta()(x)
    ['0 months', '1 month', '2 months', '3 months', '4 months']
    >>> label_timedelta(use_plurals=False)(x)
    ['0 month', '1 month', '2 month', '3 month', '4 month']
    >>> label_timedelta(units='day')(x)
    ['0 days', '31 days', '62 days', '93 days', '124 days']
    >>> label_timedelta(units='day', zero_has_units=False)(x)
    ['0', '31 days', '62 days', '93 days', '124 days']
    >>> label_timedelta(units='day', show_units=False)(x)
    ['0', '31', '62', '93', '124']
    """

    units: Optional[DurationUnit] = None
    show_units: bool = True
    zero_has_units: bool = True
    usetex: bool = False
    space: bool = True
    use_plurals: bool = True

    def __call__(self, x: NDArrayTimedelta) -> Sequence[str]:
        if len(x) == 0:
            return []

        values, units = timedelta_helper.format_info(x, self.units)
        labels = list(label_number()(values))

        if self.show_units:
            if self.usetex and units == "us":
                units = r"$\mu s$"

            if self.use_plurals and units in ("day", "week", "month", "year"):
                units_plural = f"{units}s"
            else:
                units_plural = units

            if self.space:
                units = f" {units}"
                units_plural = f" {units_plural}"
            for i, (num, label) in enumerate(zip(values, labels)):
                if num == 0 and not self.zero_has_units:
                    continue
                elif num == 1:
                    labels[i] = f"{label}{units}"
                else:
                    labels[i] = f"{label}{units_plural}"

        return labels


@dataclass
class label_pvalue:
    """
    p-values labelling

    Parameters
    ----------
    accuracy : float
        Number to round to
    add_p : bool
        Whether to prepend "p=" or "p<" to the output

    Examples
    --------
    >>> x = [.90, .15, .015, .009, 0.0005]
    >>> label_pvalue()(x)
    ['0.9', '0.15', '0.015', '0.009', '<0.001']
    >>> label_pvalue(0.1)(x)
    ['0.9', '0.1', '<0.1', '<0.1', '<0.1']
    >>> label_pvalue(0.1, True)(x)
    ['p=0.9', 'p=0.1', 'p<0.1', 'p<0.1', 'p<0.1']
    """

    accuracy: float = 0.001
    add_p: float = False

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        x = round_any(x, self.accuracy)
        below = [num < self.accuracy for num in x]

        if self.add_p:
            eq_fmt = "p={:g}".format
            below_label = f"p<{self.accuracy:g}"
        else:
            eq_fmt = "{:g}".format
            below_label = f"<{self.accuracy:g}"

        labels = [below_label if b else eq_fmt(i) for i, b in zip(x, below)]
        return labels


def ordinal(n: float, prefix="", suffix="", big_mark=""):
    # General Case: 0th, 1st, 2nd, 3rd, 4th, 5th, 6th, 7th, 8th, 9th
    # Special Case: 10th, 11th, 12th, 13th
    n = int(n)
    idx = np.min((n % 10, 4))
    _suffix = ("th", "st", "nd", "rd", "th")[idx]
    if 11 <= (n % 100) <= 13:
        _suffix = "th"

    if big_mark:
        s = f"{n:,}"
        if big_mark != ",":
            s = s.replace(",", big_mark)
    else:
        s = f"{n}"

    return f"{prefix}{s}{_suffix}{suffix}"


@dataclass
class label_ordinal:
    """
    Ordinal number labelling

    Parameters
    ----------
    prefix : str
        What to put before the value.
    suffix : str
        What to put after the value.
    big_mark : str
        The thousands separator. This is usually
        a comma or a dot.

    Examples
    --------
    >>> label_ordinal()(range(8))
    ['0th', '1st', '2nd', '3rd', '4th', '5th', '6th', '7th']
    >>> label_ordinal(suffix=' Number')(range(11, 15))
    ['11th Number', '12th Number', '13th Number', '14th Number']
    """

    prefix: str = ""
    suffix: str = ""
    big_mark: str = ""

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        labels = [
            ordinal(num, self.prefix, self.suffix, self.big_mark) for num in x
        ]
        return labels


@dataclass
class label_bytes:
    """
    Labelling byte numbers

    Parameters
    ----------
    symbol : str
        Valid symbols are "B", "kB", "MB", "GB", "TB", "PB", "EB",
        "ZB", and "YB" for SI units, and the "iB" variants for
        binary units. Default is "auto" where the symbol to be
        used is determined separately for each value of 1x.
    units : "binary" | "si"
        Which unit base to use, 1024 for "binary" or 1000 for "si".
    fmt : str, optional
        Format sting. Default is ``{:.0f}``.

    Examples
    --------
    >>> x = [1000, 1000000, 4e5]
    >>> label_bytes()(x)
    ['1000 B', '977 KiB', '391 KiB']
    >>> label_bytes(units='si')(x)
    ['1 kB', '1 MB', '400 kB']
    """

    symbol: Literal["auto"] | BytesSymbol = "auto"
    units: Literal["binary", "si"] = "binary"
    fmt: str = "{:.0f} "

    def __post_init__(self):
        if self.units == "si":
            self.base = 1000
            self._all_symbols = (
                "B",
                "kB",
                "MB",
                "GB",
                "TB",
                "PB",
                "EB",
                "ZB",
                "YB",
            )
        else:
            self.base = 1024
            self._all_symbols = (
                "B",
                "KiB",
                "MiB",
                "GiB",
                "TiB",
                "PiB",
                "EiB",
                "ZiB",
                "YiB",
            )

        # possible exponents of base: eg 1000^1, 1000^2, 1000^3, ...
        exponents = np.arange(1, len(self._all_symbols) + 1, dtype=float)
        self._powers = self.base**exponents
        self._validate_symbol(self.symbol, ("auto",) + self._all_symbols)

    def __call__(self, x: FloatArrayLike) -> Sequence[str]:
        _all_symbols = self._all_symbols
        symbol = self.symbol
        if symbol == "auto":
            power = [bisect_right(self._powers, val) for val in x]
            symbols = [_all_symbols[p] for p in power]
        else:
            power = np.array(match([symbol], _all_symbols))
            symbols = [symbol] * len(x)

        x = np.asarray(x)
        power = np.asarray(power, dtype=float)
        values = x / self.base**power
        fmt = (self.fmt + "{}").format
        labels = [fmt(v, s) for v, s in zip(values, symbols)]
        return labels

    def _validate_symbol(self, symbol: str, allowed_symbols: Sequence[str]):
        if symbol not in allowed_symbols:
            raise ValueError(
                "Symbol must be one of {}".format(allowed_symbols)
            )


# Deprecated
comma_format = label_comma
custom_format = label_custom
currency_format = label_currency
label_dollar = label_dollar
percent_format = label_percent
scientific_format = label_scientific
date_format = label_date
number_format = label_number
log_format = label_log
timedelta_format = label_timedelta
pvalue_format = label_pvalue
ordinal_format = label_ordinal
number_bytes_format = label_bytes



from __future__ import annotations

import math
import sys
from datetime import datetime, timezone
from typing import TYPE_CHECKING, cast, overload
from warnings import warn

import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

if TYPE_CHECKING:
    from datetime import tzinfo
    from typing import Any, Optional, Sequence, TypeGuard, TypeVar

    from mizani.typing import (
        AnyArrayLike,
        DurationUnit,
        FloatArrayLike,
        FloatSeries,
        NDArrayFloat,
        NullType,
        NumericUFunction,
        SeqDatetime,
        TupleFloat2,
    )

    T = TypeVar("T")

__all__ = [
    "round_any",
    "min_max",
    "match",
    "precision",
    "isclose_abs",
    "same_log10_order_of_magnitude",
    "identity",
    "get_categories",
    "get_timezone",
    "has_dtype",
]

# Use sqrt(epsilon) to correct for loss of precision due floating point
# rounding errors. This value is derived for forward difference numerical
# difference, but for our use cases this choice of value is arbitrary.
EPSILON = sys.float_info.epsilon
ROUNDING_ERROR = np.sqrt(EPSILON)
ABS_TOL = 1e-10  # Absolute Tolerance

DISCRETE_KINDS = "ObUS"
CONTINUOUS_KINDS = "ifuc"

SECONDS: dict[DurationUnit, float] = {
    "ns": 1e-9,  # nanosecond
    "us": 1e-6,  # microsecond
    "ms": 1e-3,  # millisecond
    "s": 1,  # second
    "min": 60,  # minutes
    "h": 3600,  # hour
    "day": 24 * 3600,  # day
    "week": 7 * 24 * 3600,  # week
    "month": 31 * 24 * 3600,  # month
    "year": 365 * 24 * 3600,  # year
}

NANOSECONDS: dict[DurationUnit, float] = {
    "ns": 1,  # nanosecond
    "us": 1e3,  # microsecond
    "ms": 1e6,  # millisecond
    "s": 1e9,  # second
    "min": 60e9,  # minutes
    "h": 3600e9,  # hour
    "day": 24 * 3600e9,  # day
    "week": 7 * 24 * 3600e9,  # week
    "month": 31 * 24 * 3600e9,  # month
    "year": 365 * 24 * 3600e9,  # year
}


@overload
def round_any(
    x: FloatArrayLike, accuracy: float, f: NumericUFunction = np.round
) -> NDArrayFloat: ...


@overload
def round_any(
    x: float, accuracy: float, f: NumericUFunction = np.round
) -> float: ...


def round_any(
    x: FloatArrayLike | float, accuracy: float, f: NumericUFunction = np.round
) -> NDArrayFloat | float:
    """
    Round to multiple of any number.
    """
    if not is_vector(x):
        x = np.asarray(x)
    return f(x / accuracy) * accuracy


def min_max(
    x: FloatArrayLike | float, na_rm: bool = False, finite: bool = True
) -> TupleFloat2:
    """
    Return the minimum and maximum of x

    Parameters
    ----------
    x : array_like
        Sequence
    na_rm : bool
        Whether to remove ``nan`` values.
    finite : bool
        Whether to consider only finite values.

    Returns
    -------
    out : tuple
        (minimum, maximum) of x
    """
    x = np.asarray(x)

    if na_rm and finite:
        x = x[np.isfinite(x)]
    elif not na_rm and np.any(np.isnan(x)):
        return np.nan, np.nan
    elif na_rm:
        x = x[~np.isnan(x)]
    elif finite:
        x = x[~np.isinf(x)]

    if len(x):
        return np.min(x), np.max(x)  # type: ignore
    else:
        return float("-inf"), float("inf")


def match(
    v1: AnyArrayLike,
    v2: AnyArrayLike,
    nomatch: int = -1,
    incomparables: Optional[Any] = None,
    start: int = 0,
) -> list[int]:
    """
    Return a vector of the positions of (first)
    matches of its first argument in its second.

    Parameters
    ----------
    v1: array-like
        The values to be matched

    v2: array-like
        The values to be matched against

    nomatch: int
        The value to be returned in the case when
        no match is found.

    incomparables: array-like
        A list of values that cannot be matched.
        Any value in v1 matching a value in this list
        is assigned the nomatch value.
    start: int
        Type of indexing to use. Most likely 0 or 1
    """
    # NOTE: This function gets called a lot. If it can
    # be optimised, it should.
    lookup: dict[Any, int] = {}
    for i, x in enumerate(v2, start=start):
        if x not in lookup:
            lookup[x] = i

    if incomparables:
        skip = set(incomparables)
        lst = [
            lookup.get(x, nomatch) if x not in skip else nomatch for x in v1
        ]
    else:
        lst = [lookup.get(x, nomatch) for x in v1]
    return lst


def precision(x: FloatArrayLike | float) -> float:
    """
    Return the precision of x

    Parameters
    ----------
    x : array_like | numeric
        Value(s) whose for which to compute the precision.

    Returns
    -------
    out : numeric
        The precision of ``x`` or that the values in ``x``.

    Notes
    -----
    The precision is computed in base 10.

    Examples
    --------
    >>> precision([0.08, 0.09])
    0.01

    Maximum precision is 1

    >>> precision([9, 8])
    1
    >>> precision([16, 78])
    1

    A single values have a precision of 1

    >>> precision(0.08)
    1
    >>> precision([325])
    1
    """

    x = np.asarray(x)
    x = x[~(np.isnan(x) | np.isinf(x))]

    if len(x) <= 1:
        return 1

    smallest_diff = np.diff(np.sort(x))[0]
    if smallest_diff < ROUNDING_ERROR:  # pragma: nocover
        return 1
    else:
        # For some intel processors (Skylake), numpy may be compiled
        # to do a fast but precision losing np.log10 calculation.
        # We add a rounding error incase due to lost precision, a
        # result that should be an integer (or slightly greater than
        # one) will get "floored" to the correct value
        res = 10 ** int(np.floor(np.log10(smallest_diff) + ROUNDING_ERROR) - 1)
        has_extra_zeros = (np.round(x / res) % 10 == 0).all()
        if has_extra_zeros:
            res *= 10

        # 1 comes first so that min(1, 1.0) returns an integer
        return min(1, res)


def same_log10_order_of_magnitude(x, delta=0.1):
    """
    Return true if range is approximately in same order of magnitude

    For example these sequences are in the same order of magnitude:

        - [1, 8, 5]     # [1, 10)
        - [35, 20, 80]  # [10 100)
        - [232, 730]    # [100, 1000)

    Parameters
    ----------
    x : array-like
         Values in base 10. Must be size 2 and
        ``rng[0] <= rng[1]``.
    delta : float
        Fuzz factor for approximation. It is multiplicative.
    """
    dmin = np.log10(np.min(x) * (1 - delta))
    dmax = np.log10(np.max(x) * (1 + delta))
    return np.floor(dmin) == np.floor(dmax)


def identity(param: T) -> T:
    """
    Return whatever is passed in
    """
    return param


def get_categories(x):
    """
    Return the categories of x

    Parameters
    ----------
    x : category_like
        Input Values

    Returns
    -------
    out : Index
        Categories of x
    """
    try:
        return x.cat.categories  # series
    except AttributeError:
        try:
            return x.categories  # plain categorical
        except AttributeError as err:
            raise TypeError("x is not a categorical") from err


def log(x, base):
    """
    Calculate the log of x

    Parameters
    ----------
    x : category_like
        Input Values

    base : float
        Base of logarithm

    Returns
    -------
    out : float
        Log of x
    """
    if base == 10:
        res = np.log10(x)
    elif base == 2:
        res = np.log2(x)
    elif base == np.e:
        res = np.log(x)
    else:
        res = np.log(x) / np.log(base)
    return res


def get_timezone(x: SeqDatetime) -> tzinfo | None:
    """
    Return a single timezone for the sequence of datetimes

    Returns the timezone of first item and warns if any other items
    have a different timezone
    """

    # Ref: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    x0 = next(iter(x))
    if not isinstance(x0, datetime):
        return None

    x = cast(list[datetime], x)
    info = x0.tzinfo
    if info is None:
        return timezone.utc

    # Consistency check
    tzname0 = info.tzname(x0)
    tznames = (dt.tzinfo.tzname(dt) if dt.tzinfo else None for dt in x)

    if any(tzname0 != name for name in tznames):
        msg = (
            "Dates in column have different time zones. "
            f"Choosen `{tzname0}` the time zone of the first date. "
            "To use a different time zone, create a "
            "labeller and pass the time zone."
        )
        warn(msg)
    return info


def get_null_value(x: Any) -> NullType:
    """
    Return a Null value for the type of values
    """
    from datetime import datetime, timedelta

    import pandas as pd

    x0 = next(iter(x))
    numeric_types: Sequence[type] = (
        np.int32,
        np.int64,
        np.float64,
        int,
        float,
        bool,
    )

    if pdtypes.is_object_dtype(x):
        return None
    elif isinstance(x0, numeric_types):  # type: ignore
        return float("nan")
    # pandas types subclass cypthon types, so check
    # for them first
    elif isinstance(x0, (pd.Timestamp, pd.Timedelta)):
        return type(x0)("NaT")
    elif isinstance(x0, (datetime, timedelta)):
        return None
    elif isinstance(x0, (np.datetime64, np.timedelta64)):
        return type(x0)("NaT")
    else:
        raise ValueError(
            "Cannot get a null value for type: {}".format(type(x[0]))
        )


def isclose_abs(a: float, b: float, tol: float = ABS_TOL) -> bool:
    """
    Return True if a and b are close given the absolute tolerance
    """
    return math.isclose(a, b, rel_tol=0, abs_tol=ABS_TOL)


def is_vector(x: Any) -> TypeGuard[NDArrayFloat | FloatSeries]:
    """
    Return True if x is a numpy array or a pandas series
    """
    return isinstance(x, (np.ndarray, pd.Series))


def has_dtype(
    x: Any,
) -> TypeGuard[np.ndarray | pd.Series | pd.Categorical | pd.Index]:
    """
    Return True if x has the dtype property
    """
    return hasattr(x, "dtype")



from datetime import datetime, timedelta
from types import FunctionType, MethodType
from zoneinfo import ZoneInfo

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from numpy.testing import assert_equal

from mizani.breaks import breaks_extended, minor_breaks
from mizani.transforms import (
    asn_trans,
    atanh_trans,
    boxcox_trans,
    datetime_trans,
    exp_trans,
    gettrans,
    identity_trans,
    log1p_trans,
    log2_trans,
    log10_trans,
    log_trans,
    modulus_trans,
    pd_timedelta_trans,
    probability_trans,
    pseudo_log_trans,
    reciprocal_trans,
    reverse_trans,
    sqrt_trans,
    symlog_trans,
    timedelta_trans,
    trans,
    trans_new,
)

arr = np.arange(1, 100)


def test_trans():
    with pytest.raises(TypeError):
        trans()  # type: ignore

    with pytest.raises(AttributeError):
        identity_trans(universe=True)


def test_trans_new():
    t = trans_new(
        "bounded_identity",
        staticmethod(lambda x: x),
        classmethod(lambda x: x),
        _format=lambda x: str(x),
        domain=(-999, 999),
        doc="Bounded Identity transform",
    )

    assert t.__name__ == "bounded_identity_trans"
    assert isinstance(t.transform, FunctionType)
    assert isinstance(t.inverse, MethodType)
    assert isinstance(t.format, FunctionType)
    assert t.domain == (-999, 999)
    assert t.__doc__ == "Bounded Identity transform"

    # ticks do not go beyond the bounds
    major = t().breaks((-1999, 1999))
    assert min(major) >= -999
    assert max(major) <= 999


def test_gettrans():
    t0 = identity_trans()
    t1 = gettrans(t0)
    t2 = gettrans(identity_trans)
    t3 = gettrans("identity")
    t4 = gettrans()
    assert all(isinstance(x, identity_trans) for x in (t0, t1, t2, t3, t4))

    t = gettrans(exp_trans)
    assert t.__class__.__name__ == "power_e_trans"

    with pytest.raises(ValueError):
        gettrans(object)


def _test_trans(trans, x, *args, **kwargs):
    t = gettrans(trans)
    xt = t.transform(x)
    x2 = t.inverse(xt)
    is_log_trans = "log" in t.__class__.__name__ and hasattr(t, "base")
    # round trip
    npt.assert_allclose(x, x2)
    major = t.breaks([min(x), max(x)])
    minor = t.minor_breaks(t.transform(major))
    # Breaks and they are finite
    assert len(major)
    if is_log_trans and int(t.base) == 2:
        # Minor breaks for base == 2
        assert len(minor) == 0
    else:
        assert len(minor)
    assert all(np.isfinite(major))
    assert all(np.isfinite(minor))
    # Not breaks outside the domain
    assert all(major >= t.domain[0])
    assert all(major <= t.domain[1])
    assert all(minor >= t.domain[0])
    assert all(minor <= t.domain[1])


def test_asn_trans():
    (_test_trans(asn_trans, arr * 0.01),)


def test_atanh_trans():
    (_test_trans(atanh_trans, arr * 0.001),)


def test_boxcox_trans():
    _test_trans(boxcox_trans(0.5), arr * 10)
    _test_trans(boxcox_trans(1), arr)
    with pytest.raises(ValueError):
        x = np.arange(-4, 4)
        _test_trans(boxcox_trans(0.5), x)

    # Special case, small p and p = 0
    with pytest.warns(RuntimeWarning):
        _test_trans(boxcox_trans(1e-8), arr)
        _test_trans(boxcox_trans(0), arr)

    x = [0, 1, 2, 3]
    t = boxcox_trans(0)
    with pytest.warns(RuntimeWarning):
        xt = t.transform(x)
    xti = t.inverse(xt)
    assert np.isneginf(xt[0])
    npt.assert_array_almost_equal(x, xti)


def test_modulus_trans():
    _test_trans(modulus_trans(0), arr)
    _test_trans(modulus_trans(0.5), arr * 10)


def test_exp_trans():
    _test_trans(exp_trans, arr)

    exp2_trans = exp_trans(2)
    _test_trans(exp2_trans, arr * 0.1)


def test_identity_trans():
    _test_trans(identity_trans, arr)


def test_log10_trans():
    _test_trans(log10_trans, arr)


def test_log1p_trans():
    _test_trans(log1p_trans, arr)


def test_log2_trans():
    _test_trans(log2_trans, arr)


def test_log_trans():
    _test_trans(log_trans, arr)


def test_reverse_trans():
    _test_trans(reverse_trans, arr)


def test_sqrt_trans():
    _test_trans(sqrt_trans, arr)


def test_logn_trans():
    log3_trans = log_trans(3)
    _test_trans(log3_trans, arr)

    log4_trans = log_trans(
        4,
        domain=(0.1, 100),
        breaks=breaks_extended(),
        minor_breaks=minor_breaks(),
    )
    _test_trans(log4_trans, arr)


def test_reciprocal_trans():
    x = np.arange(10, 21)
    _test_trans(reciprocal_trans, x)


def test_pseudo_log_trans():
    p = np.arange(-4, 4)
    pos = [10 ** int(x) for x in p]
    arr = np.hstack([-np.array(pos[::-1]), pos])
    _test_trans(pseudo_log_trans, arr)
    _test_trans(pseudo_log_trans(base=16), arr)


def test_symlog_trans():
    p = np.arange(-4, 4)
    pos = [10 ** int(x) for x in p]
    arr = np.hstack([-np.array(pos[::-1]), pos])
    _test_trans(symlog_trans, arr)


def test_probability_trans():
    with pytest.raises(ValueError):
        t = probability_trans("unknown_distribution")

    # cdf of the normal is centered at 0 and
    # The values either end of 0 are symmetric
    x = [-3, -2, -1, 0, 1, 2, 3]
    t = probability_trans("norm")
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert xt[3] == 0.5
    npt.assert_allclose(xt[:3], 1 - xt[-3:][::-1])
    npt.assert_allclose(x, x2)


def test_datetime_trans():
    UTC = ZoneInfo("UTC")

    x = [datetime(year, 1, 1, tzinfo=UTC) for year in [2010, 2015, 2020, 2026]]
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(a == b for a, b in zip(x, x2))

    # numpy datetime64
    x = [np.datetime64(i, "D") for i in range(1, 11)]
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(isinstance(val, datetime) for val in x2)

    # pandas timestamp
    x = pd.date_range(start="1/1/2022", end="1/2/2022", freq="3h", tz="EST")
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(x == x2)

    # Irregular index
    s = pd.Series(x, index=range(11, len(x) + 11))
    st = t.transform(s)
    s2 = t.inverse(st)
    assert all(s == s2)


def test_datetime_trans_tz():
    EST = ZoneInfo("EST")
    UTC = ZoneInfo("UTC")

    x = [datetime(2022, 1, 1, 0 + 3 * i, 0, 0, tzinfo=EST) for i in range(8)]

    # Same trans as data
    t = datetime_trans()
    x2 = t.inverse(t.transform(x))
    assert_equal(x, x2)
    assert all(val.tzinfo == EST for val in x2)

    # UTC trans
    t = datetime_trans(UTC)
    x2 = t.inverse(t.transform(x))
    assert_equal(x, x2)
    assert all(val.tzinfo == UTC for val in x2)

    t = datetime_trans("MST")
    assert t.tzinfo == t.tz
    assert_equal(t.transform([]), np.array([]))


def test_timedelta_trans():
    x = [timedelta(days=i) for i in range(1, 11)]
    t = timedelta_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(a == b for a, b in zip(x, x2))


def test_pd_timedelta_trans():
    x = [pd.Timedelta(days=i) for i in range(1, 11)]
    t = pd_timedelta_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(a == b for a, b in zip(x, x2))



from docutils.parsers.rst.roles import code_role


def python_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    options = {"language": "python"}
    return code_role(
        role, rawtext, text, lineno, inliner, options=options, content=content
    )


def setup(app):
    app.add_role("python", python_role)
    app.add_role("py", python_role)



from __future__ import annotations

import os
import re
import shlex
import sys
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Literal, Optional, Sequence, TypeAlias

TPL_FILENAME = "release-checklist-tpl.md"
THIS_DIR = Path(__file__).parent
NEW_ISSUE = "https://github.com/has2k1/mizani/issues/new"

VersionPart: TypeAlias = Literal[
    "major",
    "minor",
    "patch",
]

count = r"(?:[0-9]|[1-9][0-9]+)"
DESCRIBE_PATTERN = re.compile(
    r"^v"
    rf"(?P<version>{count}\.{count}\.{count})"
    rf"(?P<pre>(a|b|rc){count})?"
    r"(-(?P<commits>\d+)-g(?P<hash>[a-z0-9]+))?"
    r"(?P<dirty>-dirty)?"
    r"$"
)


def run(cmd: str | Sequence[str], input: Optional[str] = None) -> str:
    """
    Run command
    """
    if isinstance(cmd, str) and os.name == "posix":
        cmd = shlex.split(cmd)
    with Popen(
        cmd, stdin=PIPE, stderr=PIPE, stdout=PIPE, text=True, encoding="utf-8"
    ) as p:
        stdout, _ = p.communicate(input=input)
    return stdout.strip()


def copy_to_clipboard(s: str):
    """
    Copy s to clipboard
    """
    import platform

    plat = platform.system()

    platform_cmds = {"Darwin": "pbcopy", "Linux": "xclip", "Windows": "clip"}

    try:
        from pandas.io import clipboard
    except ImportError:
        try:
            cmd = platform_cmds[plat]
        except KeyError as err:
            msg = f"No clipboard for this system: {plat}"
            raise RuntimeError(msg) from err
        run(cmd, input=s)
    else:
        clipboard.copy(s)  # type: ignore


def get_previous_version(s: Optional[str] = None) -> str:
    """
    Get previous version

    Either the 2nd commandline arg (v) or obtained from git describe
    """
    if s:
        vtxt = s if s.startswith("v") else f"v{s}"
    else:
        cmd = "git describe --dirty --tags --long --match '*[0-9]*'"
        vtxt = run(cmd)

    m = DESCRIBE_PATTERN.match(vtxt)
    if not m:
        raise ValueError(f"Bad version: {vtxt}")

    return m.group("version")


def bump_version(version: str, part: VersionPart) -> str:
    """
    Bump version
    """
    parts = version.split(".")
    i = ("major", "minor", "patch").index(part)
    parts[i] = str(int(parts[i]) + 1)
    # Zero-out the smaller parts
    for j in range(i + 1, 3):
        parts[j] = "0"
    return ".".join(parts)


def generate_checklist(version: str) -> str:
    """
    Generate checklist for releasing the given version
    """
    path = THIS_DIR / TPL_FILENAME
    pattern = re.compile(
        # The template is everything below the dashed line
        r"\n-+\n(?P<tpl>.+)",
        flags=re.MULTILINE | re.DOTALL,
    )
    with Path(path).open("r") as f:
        contents = f.read()

    m = pattern.search(contents)
    if not m:
        raise ValueError(f"Cannot find the relevant content in '{path}'")

    tpl = m.group("tpl")
    return tpl.replace("<VERSION>", version)


def process(part: VersionPart, prev: str | None):
    """
    Run the full process

    1. Calculate the next version from the previous version
    2. Add the next version to the checklist template
    3. Copy the template to the system clipboard
    """
    prev_version = get_previous_version(prev)
    next_version = bump_version(prev_version, part)
    cl = generate_checklist(next_version)
    copy_to_clipboard(cl)
    verbose(prev_version, next_version)


def verbose(prev_version, next_version):
    """
    Print version details
    """
    from textwrap import dedent

    from term import T0 as T  # type: ignore

    s = f"""
    Previous Version: {T(prev_version, "lightblue", effect="strikethrough")}
        Next Version: {T(next_version, "lightblue", effect="bold")}

    The release checklist has been copied to the clipboard. Use it to
    open a new issue at: {T(NEW_ISSUE, "yellow")}\
    """
    print(dedent(s))


if __name__ == "__main__":
    if len(sys.argv) >= 2:
        part = sys.argv[1]
        prev = sys.argv[2] if len(sys.argv) >= 3 else None
        assert part in ("major", "minor", "patch")
        process(part, prev)



from __future__ import annotations

import math
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta

from ..utils import isclose_abs
from .types import DateFrequency

if TYPE_CHECKING:
    from mizani.typing import (
        DatetimeBreaksUnits,
        TupleDatetime2,
        TupleFloat2,
        TupleInt2,
    )


SECONDS_PER_DAY = 24 * 60 * 60
SECONDS_PER_HOUR = 60 * 60
ONE_DAY = timedelta(days=1)
ONE_WEEK = timedelta(days=7)
ONE_MONTH = relativedelta(months=1)
ONE_YEAR = relativedelta(years=1)
ONE_HOUR = timedelta(hours=1)
ONE_MINUTE = timedelta(minutes=1)
ONE_SECOND = timedelta(seconds=1)
DF = DateFrequency


@dataclass
class Interval:
    start: datetime
    end: datetime

    def __post_init__(self):
        if isinstance(self.start, date):
            self.start = datetime.fromisoformat(self.start.isoformat())

        if isinstance(self.end, date):
            self.end = datetime.fromisoformat(self.end.isoformat())

        self._delta = relativedelta(self.end, self.start)
        self._tdelta = self.end - self.start

    @property
    def y(self) -> int:
        """
        Years
        """
        return self._delta.years

    @property
    def y_wide(self) -> int:
        """
        Years (enclosing the original)
        """
        return Interval(*self.limits_year()).y

    @property
    def M(self) -> int:
        """
        Months
        """
        return self.y * 12 + self._delta.months

    @property
    def M_wide(self) -> int:
        """
        Months (enclosing the original)
        """
        return Interval(*self.limits_month()).M

    @property
    def w(self) -> int:
        """
        Weeks
        """
        return self._tdelta.days // 7

    @property
    def w_wide(self) -> int:
        """
        Weeks (enclosing the original)
        """
        return Interval(*self.limits_week()).w

    @property
    def d(self) -> int:
        """
        Days
        """
        return self._tdelta.days

    @property
    def d_wide(self) -> int:
        """
        Days (enclosing the original)
        """
        return Interval(*self.limits_day()).d

    @property
    def h(self) -> int:
        """
        Hours
        """
        return self.d * 24 + self._delta.hours

    @property
    def h_wide(self) -> int:
        """
        Hours (enclosing the original)
        """
        return Interval(*self.limits_hour()).h

    @property
    def m(self) -> int:
        """
        Minutes
        """
        return self.h * 60 + self._delta.minutes

    @property
    def m_wide(self) -> int:
        """
        Minutes (enclosing the original)
        """
        return Interval(*self.limits_minute()).m

    @property
    def s(self) -> int:
        """
        Seconds
        """
        return math.floor(self._tdelta.total_seconds())

    @property
    def u(self) -> int:
        """
        Microseconds
        """
        return math.floor(self._tdelta.total_seconds() * 1e6)

    @property
    def limits(self) -> TupleDatetime2:
        return self.start, self.end

    def limits_year(self) -> TupleDatetime2:
        return floor_year(self.start), ceil_year(self.end)

    def limits_month(self) -> TupleDatetime2:
        return round_month(self.start), round_month(self.end)

    def limits_week(self) -> TupleDatetime2:
        return floor_week(self.start), ceil_week(self.end)

    def limits_day(self) -> TupleDatetime2:
        return floor_day(self.start), ceil_day(self.end)

    def limits_hour(self) -> TupleDatetime2:
        return floor_hour(self.start), ceil_hour(self.end)

    def limits_minute(self) -> TupleDatetime2:
        return floor_minute(self.start), ceil_minute(self.end)

    def limits_second(self) -> TupleDatetime2:
        return floor_second(self.start), ceil_second(self.end)

    def limits_for_frequency(self, freq: DateFrequency) -> TupleDatetime2:
        lookup = {
            DF.YEARLY: self.limits_year,
            DF.MONTHLY: self.limits_month,
            DF.DAILY: self.limits_day,
            DF.HOURLY: self.limits_hour,
            DF.MINUTELY: self.limits_minute,
            DF.SECONDLY: self.limits_second,
        }
        try:
            return lookup[freq]()
        except KeyError:
            return self.limits


def floor_year(d: datetime) -> datetime:
    """
    Round down to the start of the year
    """
    return d.min.replace(d.year)


def ceil_year(d: datetime) -> datetime:
    """
    Round up to start of the year
    """
    _d_floor = floor_year(d)
    if d == _d_floor:
        # Already at the start of the year
        return d
    else:
        return _d_floor + ONE_YEAR


def floor_mid_year(d: datetime) -> datetime:
    """
    Round down half a year
    """
    _d_floor = floor_year(d)
    if d.month < 7:
        return _d_floor.replace(month=1)
    else:
        return _d_floor.replace(month=7)


def ceil_mid_year(d: datetime) -> datetime:
    """
    Round up half a year
    """
    _d_floor = floor_year(d)
    if d == _d_floor:
        return d
    elif d.month < 7:
        return _d_floor.replace(month=7)
    else:
        return _d_floor + ONE_YEAR


def floor_month(d: datetime) -> datetime:
    """
    Round down to the start of the month
    """
    return d.min.replace(d.year, d.month, 1)


def ceil_month(d: datetime) -> datetime:
    """
    Round up to the start of the next month
    """
    _d_floor = floor_month(d)
    if d == _d_floor:
        return d

    return _d_floor + ONE_MONTH


def round_month(d: datetime) -> datetime:
    """
    Round date to the "nearest" start of the month
    """
    return floor_month(d) if d.day <= 15 else ceil_month(d)


def floor_week(d: datetime) -> datetime:
    """
    Round down to the start of the week

    Week start on are on 1st, 8th, 15th and 22nd
    day of the month
    """
    if d.day < 8:
        day = 1
    elif d.day < 15:
        day = 8
    elif d.day < 22:
        day = 15
    else:
        day = 22
    return d.min.replace(d.year, d.month, day)


def ceil_week(d: datetime) -> datetime:
    """
    Round up to the start of the next month

    Week start on are on 1st, 8th, 15th and 22nd
    day of the month
    """
    _d_floor = floor_week(d)
    if d == _d_floor:
        return d

    if d.day >= 22:
        return d.min.replace(d.year, d.month, d.day) + ONE_WEEK

    return _d_floor + ONE_WEEK


def floor_day(d: datetime) -> datetime:
    """
    Round down to the start of the day
    """
    return (
        d.min.replace(d.year, d.month, d.day, tzinfo=d.tzinfo)
        if has_time(d)
        else d
    )


def ceil_day(d: datetime) -> datetime:
    """
    Round up to the start of the next day
    """
    return floor_day(d) + ONE_DAY if has_time(d) else d


def floor_hour(d: datetime) -> datetime:
    """
    Round down to the start of the hour
    """
    if at_the_hour(d):
        return d
    return floor_day(d).replace(hour=d.hour)


def ceil_hour(d: datetime) -> datetime:
    """
    Round up to the start of the next hour
    """
    if at_the_hour(d):
        return d
    return floor_hour(d) + ONE_HOUR


def floor_minute(d: datetime) -> datetime:
    """
    Round down to the start of the minute
    """
    if at_the_minute(d):
        return d
    return floor_hour(d).replace(minute=d.minute)


def ceil_minute(d: datetime) -> datetime:
    """
    Round up to the start of the next minute
    """
    if at_the_minute(d):
        return d
    return floor_minute(d) + ONE_MINUTE


def floor_second(d: datetime) -> datetime:
    """
    Round down to the start of the second
    """
    if at_the_second(d):
        return d
    return floor_minute(d).replace(second=d.second)


def ceil_second(d: datetime) -> datetime:
    """
    Round up to the start of the next minute
    """
    if at_the_second(d):
        return d
    return floor_second(d) + ONE_SECOND


def has_time(d: datetime) -> bool:
    """
    Return True if the time of datetime is not 00:00:00 (midnight)
    """
    return d.time() != time.min


def at_the_hour(d: datetime) -> bool:
    """
    Return True if the time of datetime is at the hour mark
    """
    t = d.time()
    return t.minute == 0 and t.second == 0 and t.microsecond == 0


def at_the_minute(d: datetime) -> bool:
    """
    Return True if the time of datetime is at the minute mark
    """
    t = d.time()
    return t.second == 0 and t.microsecond == 0


def at_the_second(d: datetime) -> bool:
    """
    Return True if the time of datetime is at the second mark
    """
    return d.time().microsecond == 0


def align_limits(limits: TupleInt2, width: float) -> TupleFloat2:
    """
    Return limits so that breaks should be multiples of the width

    The new limits are equal or contain the original limits
    """
    low, high = limits

    l, m = divmod(low, width)
    if isclose_abs(m / width, 1):
        l += 1

    h, m = divmod(high, width)
    if not isclose_abs(m / width, 0):
        h += 1

    return l * width, h * width


def shift_limits_down(
    candidate_limits: TupleInt2, original_limits: TupleInt2, width: int
) -> TupleInt2:
    """
    Shift candidate limits down so that they can be a multiple of width

    If the shift would exclude any of the original_limits (high),
    candidate limits are returned.

    The goal of this function is to convert abitrary limits into "nicer"
    ones.
    """
    low, high = candidate_limits
    low_orig, high_orig = original_limits

    l, m = divmod(low, width)
    if isclose_abs(m / width, 1):  # pragma: no cover
        l += 1

    low_new = l * width
    diff = low - low_new
    high_new = high - diff

    if high_orig <= high_new:
        return low_new, high_new

    return candidate_limits


def expand_datetime_limits(
    limits: TupleDatetime2, width: int, units: DatetimeBreaksUnits
) -> TupleDatetime2:
    ival = Interval(*limits)
    if units == "year":
        start, end = ival.limits_year()
        span = ival.y_wide
    elif units == "month":
        start, end = ival.limits_month()
        span = ival.M_wide
    elif units == "week":
        start, end = ival.limits_week()
        span = ival.w_wide
    elif units == "day":
        start, end = ival.limits_day()
        span = ival.d_wide
    elif units == "hour":
        start, end = ival.limits_hour()
        span = ival.h_wide
    elif units == "minute":
        start, end = ival.limits_minute()
        span = ival.m_wide
    elif units == "second":
        start, end = ival.limits_second()
        span = ival.s
    else:
        return limits

    new_span = math.ceil(span / width) * width
    if units == "week":
        units, new_span = "day", new_span * 7

    end = start + relativedelta(None, None, **{f"{units}s": new_span})

    if units == "year":
        limits_orig = limits[0].year, limits[1].year
        y1, y2 = shift_limits_down((start.year, end.year), limits_orig, width)
        start = start.replace(y1)
        end = end.replace(y2)

    return start, end



"""
Check the Tag on Github Actions

Writes the following variables to the output file for the github step:

    1. tag_type=<tag_type>
       Possible values are:
           - alpha
           - beta
           - candidate
           - final
           - ""

    2. publish_on=<publish_on>
       Possible values are:
           - pypi
           - testpypi
           - ""

The output of this script should tell you whether to publish a release
and to what index.
"""

import os
import re
from pathlib import Path
from typing import Literal, TypeAlias

TagType: TypeAlias = Literal[
    "alpha",
    "beta",
    "candidate",
    "final",
]


count = r"(?:[0-9]|[1-9][0-9]+)"
SEMVER_PATTERN = re.compile(
    r"^v"
    rf"(?P<version>{count}\.{count}\.{count})"
    rf"(?P<pre>(a|b|rc){count})?"
    r"$"
)


def get_tag_type() -> TagType | None:
    """
    Return the type of semantic version
    """
    tag = os.environ.get("GITHUB_REF_NAME")

    # Probably not on GHA
    if not tag:
        return

    m = SEMVER_PATTERN.match(tag)
    if not m:
        return

    pre = m.group("pre")

    if pre:
        if pre == "a":
            return "alpha"
        elif pre == "b":
            return "beta"
        else:
            return "candidate"

    return "final"


def output_tag_type(tag_type: TagType | None):
    """
    Write tag_type to the GHA output
    """
    if not (output_file := os.environ.get("GITHUB_OUTPUT")):
        return

    with Path(output_file).open("a") as f:
        print(f"tag_type={tag_type}", file=f)


def output_publish_on(tag_type: TagType | None):
    """
    Write index (pypi or testpypi) to publish on to the GHA output

    i.e. Where to release
    """
    if not (output_file := os.environ.get("GITHUB_OUTPUT")):
        return

    # Probably not on GHA
    if not output_file:
        return

    if tag_type in ("alpha", "beta", "candidate"):
        publish_on = "testpypi"
    elif tag_type == "final":
        publish_on = "pypi"
    else:
        publish_on = ""

    with Path(output_file).open("a") as f:
        print(f"publish_on={publish_on}", file=f)


if __name__ == "__main__":
    tag_type = get_tag_type()
    output_tag_type(tag_type)
    output_publish_on(tag_type)



"""
All scales have a means by which the values that are mapped
onto the scale are interpreted. Numeric digital scales put
out numbers for direct interpretation, but most scales
cannot do this. What they offer is named markers/ticks that
aid in assessing the values e.g. the common odometer will
have ticks and values to help gauge the speed of the vehicle.

The named markers are what we call breaks. Properly calculated
breaks make interpretation straight forward. These functions
provide ways to calculate good(hopefully) breaks.
"""

from __future__ import annotations

import sys
from datetime import datetime, timedelta
from itertools import product
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from mizani._core.dates import (
    calculate_date_breaks_auto,
    calculate_date_breaks_byunits,
)

from .utils import NANOSECONDS, SECONDS, log, min_max

if TYPE_CHECKING:
    from typing import Callable, Literal, Optional, Sequence

    from mizani.typing import (
        DatetimeBreaksUnits,
        DurationUnit,
        FloatArrayLike,
        NDArrayFloat,
        NDArrayTimedelta,
        Timedelta,
        Trans,
        TupleFloat2,
        TupleFloat5,
        TupleT2,
    )


__all__ = [
    "breaks_log",
    "breaks_symlog",
    "minor_breaks",
    "minor_breaks_trans",
    "breaks_date",
    "breaks_timedelta",
    "breaks_extended",
]


class breaks_log:
    """
    Integer breaks on log transformed scales

    Parameters
    ----------
    n : int
        Desired number of breaks
    base : int
        Base of logarithm

    Examples
    --------
    >>> x = np.logspace(3, 6)
    >>> limits = min(x), max(x)
    >>> breaks_log()(limits)
    array([     1000,    10000,   100000,  1000000])
    >>> breaks_log(2)(limits)
    array([  1000, 100000])
    >>> breaks_log()([0.1, 1])
    array([0.1, 0.3, 1. , 3. ])
    """

    def __init__(self, n: int = 5, base: float = 10):
        self.n = n
        self.base = base

    def __call__(self, limits: TupleFloat2) -> NDArrayFloat:
        """
        Compute breaks

        Parameters
        ----------
        limits : tuple
            Minimum and maximum values

        Returns
        -------
        out : array_like
            Sequence of breaks points
        """
        if any(np.isinf(limits)):
            return np.array([])

        n = self.n
        base = self.base
        rng = log(limits, base)
        _min = int(np.floor(rng[0]))
        _max = int(np.ceil(rng[1]))

        # Prevent overflow
        if float(base) ** _max > sys.maxsize:
            base = float(base)

        if _max == _min:
            return np.array([base**_min])

        # Try getting breaks at the integer powers of the base
        # e.g [1, 100, 10000, 1000000]
        # If there are too few breaks, try other points using the
        # _log_sub_breaks
        by = int(np.floor((_max - _min) / n)) + 1
        for step in range(by, 0, -1):
            breaks = np.array([base**i for i in range(_min, _max + 1, step)])
            relevant_breaks = (limits[0] <= breaks) & (breaks <= limits[1])
            if np.sum(relevant_breaks) >= n - 2:
                return breaks

        return _breaks_log_sub(n=n, base=base)(limits)


class _breaks_log_sub:
    """
    Breaks for log transformed scales

    Calculate breaks that do not fall on integer powers of
    the base.

    Parameters
    ----------
    n : int
        Desired number of breaks
    base : int | float
        Base of logarithm

    Notes
    -----
    Credit: Thierry Onkelinx (thierry.onkelinx@inbo.be) for the original
    algorithm in the r-scales package.
    """

    def __init__(self, n: int = 5, base: float = 10):
        self.n = n
        self.base = base

    def __call__(self, limits: TupleFloat2) -> NDArrayFloat:
        base = self.base
        n = self.n
        rng = log(limits, base)
        _min = int(np.floor(rng[0]))
        _max = int(np.ceil(rng[1]))
        steps = [1]

        # Prevent overflow
        if float(base) ** _max > sys.maxsize:
            base = float(base)

        def delta(x):
            """
            Calculates the smallest distance in the log scale between the
            currently selected breaks and a new candidate 'x'
            """
            arr = np.sort(np.hstack([x, steps, base]))
            log_arr = log(arr, base)
            return np.min(np.diff(log_arr))

        if self.base == 2:
            return np.array([base**i for i in range(_min, _max + 1)])

        candidate = np.arange(base + 1)
        candidate = np.compress(
            (candidate > 1) & (candidate < base), candidate
        )

        while len(candidate):
            best = np.argmax([delta(x) for x in candidate])
            steps.append(candidate[best])
            candidate = np.delete(candidate, best)
            _factors = [base**i for i in range(_min, _max + 1)]
            breaks = np.array([f * s for f, s in product(_factors, steps)])
            relevant_breaks = (limits[0] <= breaks) & (breaks <= limits[1])

            if np.sum(relevant_breaks) >= n - 2:
                breaks = np.sort(breaks)
                lower_end = np.max(
                    [
                        np.min(np.where(limits[0] <= breaks)) - 1,
                        0,  # type: ignore
                    ]
                )
                upper_end = np.min(
                    [
                        np.max(np.where(breaks <= limits[1])) + 1,
                        len(breaks),  # type: ignore
                    ]
                )
                return breaks[lower_end : upper_end + 1]
        else:
            return breaks_extended(n=n)(limits)


class minor_breaks:
    """
    Compute minor breaks

    This is the naive method. It does not take into account
    the transformation.

    Parameters
    ----------
    n : int
        Number of minor breaks between the major
        breaks.

    Examples
    --------
    >>> major = [1, 2, 3, 4]
    >>> limits = [0, 5]
    >>> minor_breaks()(major, limits)
    array([0.5, 1.5, 2.5, 3.5, 4.5])
    >>> minor_breaks()([1, 2], (1, 2))
    array([1.5])

    More than 1 minor break.

    >>> minor_breaks(3)([1, 2], (1, 2))
    array([1.25, 1.5 , 1.75])
    >>> minor_breaks()([1, 2], (1, 2), 3)
    array([1.25, 1.5 , 1.75])
    """

    def __init__(self, n: int = 1):
        self.n = n

    def __call__(
        self,
        major: FloatArrayLike,
        limits: Optional[TupleFloat2] = None,
        n: Optional[int] = None,
    ) -> NDArrayFloat:
        """
        Minor breaks

        Parameters
        ----------
        major : array_like
            Major breaks
        limits : array_like | None
            Limits of the scale. If *array_like*, must be
            of size 2. If **None**, then the minimum and
            maximum of the major breaks are used.
        n : int
            Number of minor breaks between the major
            breaks. If **None**, then *self.n* is used.

        Returns
        -------
        out : array_like
            Minor beraks
        """
        if len(major) < 2:
            return np.array([])

        if limits is None:
            low, high = min_max(major)
        else:
            low, high = min_max(limits)

        if n is None:
            n = self.n

        # Try to infer additional major breaks so that
        # minor breaks can be generated beyond the first
        # and last major breaks
        diff = np.diff(major)
        step = diff[0]
        if len(diff) > 1 and all(diff == step):
            major = np.hstack([major[0] - step, major, major[-1] + step])

        mbreaks = []
        factors = np.arange(1, n + 1)
        for lhs, rhs in zip(major[:-1], major[1:]):
            sep = (rhs - lhs) / (n + 1)
            mbreaks.append(lhs + factors * sep)

        minor = np.hstack(mbreaks)
        minor = minor.compress((low <= minor) & (minor <= high))
        return minor


class minor_breaks_trans:
    """
    Compute minor breaks for transformed scales

    The minor breaks are computed in data space.
    This together with major breaks computed in
    transform space reveals the non linearity of
    of a scale. See the log transforms created
    with :func:`log_trans` like :class:`log10_trans`.

    Parameters
    ----------
    trans : trans or type
        Trans object or trans class.
    n : int
        Number of minor breaks between the major
        breaks.

    Examples
    --------
    >>> from mizani.transforms import sqrt_trans
    >>> major = [1, 2, 3, 4]
    >>> limits = [0, 5]
    >>> t1 = sqrt_trans()
    >>> t1.minor_breaks(major, limits)
    array([1.58113883, 2.54950976, 3.53553391])

    # Changing the regular `minor_breaks` method

    >>> t2 = sqrt_trans()
    >>> t2.minor_breaks = minor_breaks()
    >>> t2.minor_breaks(major, limits)
    array([0.5, 1.5, 2.5, 3.5, 4.5])

    More than 1 minor break

    >>> major = [1, 10]
    >>> limits = [1, 10]
    >>> t2.minor_breaks(major, limits, 4)
    array([2.8, 4.6, 6.4, 8.2])
    """

    def __init__(self, trans: Trans, n: int = 1):
        self.trans = trans
        self.n = n

    def __call__(
        self,
        major: FloatArrayLike,
        limits: Optional[TupleFloat2] = None,
        n: Optional[int] = None,
    ) -> NDArrayFloat:
        """
        Minor breaks for transformed scales

        Parameters
        ----------
        major : array_like
            Major breaks
        limits : array_like | None
            Limits of the scale. If *array_like*, must be
            of size 2. If **None**, then the minimum and
            maximum of the major breaks are used.
        n : int
            Number of minor breaks between the major
            breaks. If **None**, then *self.n* is used.

        Returns
        -------
        out : array_like
            Minor breaks
        """
        if limits is None:
            limits = min_max(major)

        if n is None:
            n = self.n

        major = self._extend_breaks(major)
        major = self.trans.inverse(major)
        limits = self.trans.inverse(limits)
        minor = minor_breaks(n)(major, limits)
        return self.trans.transform(minor)

    def _extend_breaks(self, major: FloatArrayLike) -> FloatArrayLike:
        """
        Append 2 extra breaks at either end of major

        If breaks of transform space are non-equidistant,
        :func:`minor_breaks` add minor breaks beyond the first
        and last major breaks. The solutions is to extend those
        breaks (in transformed space) before the minor break call
        is made. How the breaks depends on the type of transform.
        """
        trans = self.trans
        trans = trans if isinstance(trans, type) else trans.__class__
        # so far we are only certain about this extending stuff
        # making sense for log transform
        is_log = trans.__name__.startswith("log")
        diff = np.diff(major)
        step = diff[0]
        if is_log and all(diff == step):
            major = np.hstack([major[0] - step, major, major[-1] + step])
        return major


class breaks_date:
    """
    Regularly spaced dates

    Parameters
    ----------
    n :
        Desired number of breaks.
    width : str | None
        An interval specification. Must be one of
        [second, minute, hour, day, week, month, year]
        If ``None``, the interval automatic.

    Examples
    --------
    >>> from datetime import datetime
    >>> limits = (datetime(2010, 1, 1), datetime(2026, 1, 1))

    Default breaks will be regularly spaced but the spacing
    is automatically determined

    >>> breaks = breaks_date(9)
    >>> [d.year for d in breaks(limits)]
    [2010, 2012, 2014, 2016, 2018, 2020, 2022, 2024, 2026]

    Breaks at 4 year intervals

    >>> breaks = breaks_date('4 year')
    >>> [d.year for d in breaks(limits)]
    [2010, 2014, 2018, 2022, 2026]
    """

    n: int
    width: Optional[int] = None
    units: Optional[DatetimeBreaksUnits] = None

    def __init__(self, n: int = 5, width: Optional[str] = None):
        if isinstance(n, str):
            width = n

        self.n = n

        if width:
            # Parse the width specification
            # e.g. '10 months' => (10, month)
            _w, units = width.strip().lower().split()
            self.width = int(_w)
            self.units = units.rstrip("s")  # type: ignore

    def __call__(self, limits: TupleT2[datetime]) -> Sequence[datetime]:
        """
        Compute breaks

        Parameters
        ----------
        limits : tuple
            Minimum and maximum :class:`datetime.datetime` values.

        Returns
        -------
        out : array_like
            Sequence of break points.
        """
        if any(pd.isna(x) for x in limits):
            return []

        if isinstance(limits[0], np.datetime64) and isinstance(
            limits[1], np.datetime64
        ):
            limits = limits[0].astype(object), limits[1].astype(object)

        if self.units and self.width:
            return calculate_date_breaks_byunits(
                limits, self.units, self.width
            )
        else:
            return calculate_date_breaks_auto(limits, self.n)


class breaks_timedelta:
    """
    Timedelta breaks

    Returns
    -------
    out : callable ``f(limits)``
        A function that takes a sequence of two
        :class:`datetime.timedelta` values and returns
        a sequence of break points.

    Examples
    --------
    >>> from datetime import timedelta
    >>> breaks = breaks_timedelta()
    >>> x = [timedelta(days=i*365) for i in range(25)]
    >>> limits = min(x), max(x)
    >>> major = breaks(limits)
    >>> [val.total_seconds()/(365*24*60*60)for val in major]
    [0.0, 5.0, 10.0, 15.0, 20.0, 25.0]
    """

    _calculate_breaks: Callable[[TupleFloat2], NDArrayFloat]

    def __init__(self, n: int = 5, Q: Sequence[float] = (1, 2, 5, 10)):
        self._calculate_breaks = breaks_extended(n=n, Q=Q)

    def __call__(
        self, limits: tuple[Timedelta, Timedelta]
    ) -> NDArrayTimedelta:
        """
        Compute breaks

        Parameters
        ----------
        limits : tuple
            Minimum and maximum :class:`datetime.timedelta` values.

        Returns
        -------
        out : array_like
            Sequence of break points.
        """
        if any(pd.isna(x) for x in limits):
            return np.array([])

        helper = timedelta_helper(limits)
        scaled_limits = helper.scaled_limits()
        scaled_breaks = self._calculate_breaks(scaled_limits)
        breaks = helper.numeric_to_timedelta(scaled_breaks)
        return breaks


# This could be cleaned up, state overload?
class timedelta_helper:
    """
    Helper for computing timedelta breaks
    and labels.

    How to use - breaks?

    1. Initialise with a timedelta sequence/limits.
    2. Get the scaled limits and use those to calculate
       breaks using a general purpose breaks calculating
       routine. The scaled limits are in numerical format.
    3. Convert the computed breaks from numeric into timedelta.

    See, :class:`breaks_timedelta`

    How to use - formating?

    1. Call :meth:`format_info` with the timedelta values to be
       formatted and get back a tuple of numeric values and
       the units for those values.
    2. Format the values with a general purpose formatting
       routing.

    See, :class:`~mizani.labels.label_timedelta`
    """

    x: NDArrayTimedelta | Sequence[Timedelta]
    units: DurationUnit
    limits: TupleFloat2
    package: Literal["pandas", "cpython"]
    factor: float

    def __init__(
        self,
        x: NDArrayTimedelta | Sequence[Timedelta],
        units: Optional[DurationUnit] = None,
    ):
        self.x = x
        self.package = self.determine_package(x[0])
        _limits = min(x), max(x)
        self.limits = self.value(_limits[0]), self.value(_limits[1])
        self.units = units or self.best_units(_limits)
        self.factor = self.get_scaling_factor(self.units)

    @classmethod
    def determine_package(cls, td: Timedelta) -> Literal["pandas", "cpython"]:
        if hasattr(td, "components"):
            package = "pandas"
        elif hasattr(td, "total_seconds"):
            package = "cpython"
        else:
            msg = f"{td.__class__} format not yet supported."
            raise ValueError(msg)
        return package

    @classmethod
    def format_info(
        cls, x: NDArrayTimedelta, units: Optional[DurationUnit] = None
    ) -> tuple[NDArrayFloat, DurationUnit]:
        helper = cls(x, units)
        return helper.timedelta_to_numeric(x), helper.units

    def best_units(
        self, x: NDArrayTimedelta | Sequence[Timedelta]
    ) -> DurationUnit:
        """
        Determine good units for representing a sequence of timedeltas
        """
        # Read
        #   [(0.9, 's'),
        #    (9, 'm)]
        # as, break ranges between 0.9 seconds (inclusive)
        # and 9 minutes are represented in seconds. And so on.
        ts_range = self.value(max(x)) - self.value(min(x))
        package = self.determine_package(x[0])
        if package == "pandas":
            cuts: list[tuple[float, DurationUnit]] = [
                (0.9, "us"),
                (0.9, "ms"),
                (0.9, "s"),
                (9, "min"),
                (6, "h"),
                (4, "day"),
                (4, "week"),
                (4, "month"),
                (3, "year"),
            ]
            denomination = NANOSECONDS
            base_units = "ns"
        else:
            cuts = [
                (0.9, "s"),
                (9, "min"),
                (6, "h"),
                (4, "day"),
                (4, "week"),
                (4, "month"),
                (3, "year"),
            ]
            denomination = SECONDS
            base_units = "ms"

        for size, units in reversed(cuts):
            if ts_range >= size * denomination[units]:
                return units

        return base_units

    @staticmethod
    def value(td: Timedelta) -> float:
        """
        Return the numeric value representation on a timedelta
        """
        if isinstance(td, pd.Timedelta):
            return td.value
        else:
            return td.total_seconds()

    def scaled_limits(self) -> TupleFloat2:
        """
        Minimum and Maximum to use for computing breaks
        """
        _min = self.limits[0] / self.factor
        _max = self.limits[1] / self.factor
        return _min, _max

    def timedelta_to_numeric(
        self, timedeltas: NDArrayTimedelta
    ) -> NDArrayFloat:
        """
        Convert sequence of timedelta to numerics
        """
        return np.array([self.to_numeric(td) for td in timedeltas])

    def numeric_to_timedelta(self, values: NDArrayFloat) -> NDArrayTimedelta:
        """
        Convert sequence of numerical values to timedelta
        """
        if self.package == "pandas":
            return np.array(
                [pd.Timedelta(int(x * self.factor), unit="ns") for x in values]
            )
        else:
            return np.array(
                [timedelta(seconds=x * self.factor) for x in values]
            )

    def get_scaling_factor(self, units):
        if self.package == "pandas":
            return NANOSECONDS[units]
        else:
            return SECONDS[units]

    def to_numeric(self, td: Timedelta) -> float:
        """
        Convert timedelta to a number corresponding to the
        appropriate units. The appropriate units are those
        determined with the object is initialised.
        """
        if isinstance(td, pd.Timedelta):
            return td.value / NANOSECONDS[self.units]
        else:
            return td.total_seconds() / SECONDS[self.units]


class breaks_extended:
    """
    An extension of Wilkinson's tick position algorithm

    Parameters
    ----------
    n : int
        Desired number of breaks
    Q : list
        List of nice numbers
    only_inside : bool
        If ``True``, then all the breaks will be within the given
        range.
    w : list
        Weights applied to the four optimization components
        (simplicity, coverage, density, and legibility). They
        should add up to 1.

    Examples
    --------
    >>> limits = (0, 9)
    >>> breaks_extended()(limits)
    array([  0. ,   2.5,   5. ,   7.5,  10. ])
    >>> breaks_extended(n=6)(limits)
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    References
    ----------
    - Talbot, J., Lin, S., Hanrahan, P. (2010) An Extension of
      Wilkinson's Algorithm for Positioning Tick Labels on Axes,
      InfoVis 2010.

    Additional Credit to Justin Talbot on whose code this
    implementation is almost entirely based.
    """

    def __init__(
        self,
        n: int = 5,
        Q: Sequence[float] = (1, 5, 2, 2.5, 4, 3),
        only_inside: bool = False,
        w: Sequence[float] = (0.25, 0.2, 0.5, 0.05),
    ):
        self.Q = Q
        self.only_inside = only_inside
        self.w = w
        self.n = n
        # Used for lookups during the computations
        self.Q_index = {q: i for i, q in enumerate(Q)}

    def coverage(
        self, dmin: float, dmax: float, lmin: float, lmax: float
    ) -> float:
        p1 = (dmax - lmax) ** 2
        p2 = (dmin - lmin) ** 2
        p3 = (0.1 * (dmax - dmin)) ** 2
        return 1 - 0.5 * (p1 + p2) / p3

    def coverage_max(self, dmin: float, dmax: float, span: float) -> float:
        range = dmax - dmin
        if span > range:
            half = (span - range) / 2.0
            return 1 - (half**2) / (0.1 * range) ** 2
        else:
            return 1

    def density(
        self, k: float, dmin: float, dmax: float, lmin: float, lmax: float
    ) -> float:
        r = (k - 1.0) / (lmax - lmin)
        rt = (self.n - 1) / (max(lmax, dmax) - min(lmin, dmin))
        return 2 - max(r / rt, rt / r)

    def density_max(self, k: float) -> float:
        if k >= self.n:
            return 2 - (k - 1.0) / (self.n - 1.0)
        else:
            return 1

    def simplicity(
        self, q: float, j: float, lmin: float, lmax: float, lstep: float
    ) -> float:
        eps = 1e-10
        n = len(self.Q)
        i = self.Q_index[q] + 1

        if (
            (lmin % lstep < eps or (lstep - lmin % lstep) < eps)
            and lmin <= 0
            and lmax >= 0
        ):
            v = 1
        else:
            v = 0
        return (n - i) / (n - 1.0) + v - j

    def simplicity_max(self, q: float, j: float) -> float:
        n = len(self.Q)
        i = self.Q_index[q] + 1
        v = 1
        return (n - i) / (n - 1.0) + v - j

    def legibility(self, lmin: float, lmax: float, lstep: float) -> float:
        # Legibility depends on fontsize, rotation, overlap ... i.e.
        # it requires drawing or simulating drawn breaks then calculating
        # a score. Return 1 ignores all that.
        return 1

    def __call__(self, limits: TupleFloat2) -> NDArrayFloat:
        """
        Calculate the breaks

        Parameters
        ----------
        limits : array
            Minimum and maximum values.

        Returns
        -------
        out : array_like
            Sequence of break points.
        """
        Q = self.Q
        w = self.w
        only_inside = self.only_inside
        simplicity_max = self.simplicity_max
        density_max = self.density_max
        coverage_max = self.coverage_max
        simplicity = self.simplicity
        coverage = self.coverage
        density = self.density
        legibility = self.legibility
        log10 = np.log10
        ceil = np.ceil
        floor = np.floor
        # casting prevents the typechecker from mixing
        # float & np.float32
        dmin, dmax = float(limits[0]), float(limits[1])

        if dmin > dmax:
            dmin, dmax = dmax, dmin
        elif dmin == dmax:
            return np.array([dmin])

        best_score = -2.0
        best: TupleFloat5 = (0, 0, 0, 0, 0)  # Gives Empty breaks
        j = 1.0

        while j < float("inf"):
            for q in Q:
                sm = simplicity_max(q, j)

                if w[0] * sm + w[1] + w[2] + w[3] < best_score:
                    j = float("inf")
                    break

                k = 2.0
                while k < float("inf"):
                    dm = density_max(k)

                    if w[0] * sm + w[1] + w[2] * dm + w[3] < best_score:
                        break

                    delta = (dmax - dmin) / (k + 1) / j / q
                    z: float = ceil(log10(delta))

                    while z < float("inf"):
                        step = j * q * (10**z)
                        cm = coverage_max(dmin, dmax, step * (k - 1))

                        if (
                            w[0] * sm + w[1] * cm + w[2] * dm + w[3]
                            < best_score
                        ):
                            break

                        min_start = int(floor(dmax / step) * j - (k - 1) * j)
                        max_start = int(ceil(dmin / step) * j)

                        if min_start > max_start:
                            z = z + 1
                            break

                        for start in range(min_start, max_start + 1):
                            lmin = start * (step / j)
                            lmax = lmin + step * (k - 1)
                            lstep = step

                            s = simplicity(q, j, lmin, lmax, lstep)
                            c = coverage(dmin, dmax, lmin, lmax)
                            d = density(k, dmin, dmax, lmin, lmax)
                            l = legibility(lmin, lmax, lstep)

                            score = w[0] * s + w[1] * c + w[2] * d + w[3] * l

                            if score > best_score and (
                                not only_inside
                                or (lmin >= dmin and lmax <= dmax)
                            ):
                                best_score = score
                                best = (lmin, lmax, lstep, q, k)
                        z = z + 1
                    k = k + 1
            j = j + 1

        locs = best[0] + np.arange(best[4]) * best[2]
        return locs


class breaks_symlog:
    """
    Breaks for the Symmetric Logarithm Transform

    Parameters
    ----------
    n : int
        Desired number of breaks
    base : int
        Base of logarithm

    Examples
    --------
    >>> limits = (-100, 100)
    >>> breaks_symlog()(limits)
    array([-100,  -10,    0,   10,  100])
    """

    def __call__(self, limits: TupleFloat2) -> NDArrayFloat:
        def _signed_log10(x):
            return np.round(np.sign(x) * np.log10(np.sign(x) * x)).astype(int)

        l, h = _signed_log10(limits)
        exps = np.arange(l, h + 1, 1)
        return np.sign(exps) * (10 ** np.abs(exps))


# Deprecated
log_breaks = breaks_log
trans_minor_breaks = minor_breaks_trans
date_breaks = breaks_date
timedelta_breaks = breaks_timedelta
extended_breaks = breaks_extended



"""
This module has been deprecated in favour of :mod:`~mizani.labels`
"""

from .labels import (
    label_bytes,
    label_comma,
    label_currency,
    label_custom,
    label_date,
    label_dollar,
    label_log,
    label_number,
    label_ordinal,
    label_percent,
    label_pvalue,
    label_scientific,
    label_timedelta,
)

# Deprecated
comma_format = label_comma
custom_format = label_custom
currency_format = label_currency
label_dollar = label_dollar
percent_format = label_percent
scientific_format = label_scientific
date_format = label_date
number_format = label_number
log_format = label_log
timedelta_format = label_timedelta
pvalue_format = label_pvalue
ordinal_format = label_ordinal
number_bytes_format = label_bytes



from __future__ import annotations

import abc
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from typing import Sequence

    from mizani.typing import FloatArrayLike, RGBHexColor

__all__ = ("ColorMap",)


class ColorMapKind(Enum):
    sequential = "sequential"
    diverging = "diverging"
    qualitative = "qualitative"
    cyclic = "cyclic"
    miscellaneous = "miscellaneous"


class ColorMap(abc.ABC):
    """
    Base color for all color maps
    """

    @abc.abstractmethod
    def _generate_colors(self, x: FloatArrayLike) -> Sequence[RGBHexColor]:
        """
        Method to map [0, 1] values onto the a color range

        Subclasses must implement this method

        Parameters
        ----------
        x :
            Values in the range [0, 1]. O maps to the start of the
            gradient, and 1 to the end of the gradient.
        """

    def discrete_palette(self, n: int) -> Sequence[RGBHexColor]:
        """
        Return n colors from the gradient

        Subclasses can override this method

        Parameters
        ----------
        n :
            Number of colors to return from the gradient.
        """
        x = np.linspace(0, 1, n)
        return self._generate_colors(x)

    def continuous_palette(
        self, x: FloatArrayLike
    ) -> Sequence[RGBHexColor | None]:
        """
        Return colors correspondsing to proportions in x

        Parameters
        ----------
        x :
            Values in the range [0, 1]. O maps to the start of the
            gradient, and 1 to the end of the gradient.
        """
        x = np.asarray(x)
        bad_bool_idx = np.isnan(x) | np.isinf(x)
        has_bad = bad_bool_idx.any()

        if has_bad:
            x[bad_bool_idx] = 0

        hex_colors = self._generate_colors(x)

        if has_bad:
            hex_colors = [
                None if isbad else c
                for c, isbad in zip(hex_colors, bad_bool_idx)
            ]
        return hex_colors



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from ..hsluv import rgb_to_hex
from ._colormap import ColorMap, ColorMapKind

if TYPE_CHECKING:
    from typing import ClassVar, Sequence

    from mizani.typing import (
        FloatArrayLike,
        RGBHexColor,
    )

__all__ = ("CubeHelixMap",)

rotation_matrix = np.array(
    [[-0.14861, +1.78277], [-0.29227, -0.90649], [+1.97294, 0.0]]
)


@dataclass
class CubeHelixMap(ColorMap):
    start: int = 0
    rotation: float = 0.4
    gamma: float = 1.0
    hue: float = 0.8
    light: float = 0.85
    dark: float = 0.15
    reverse: bool = False

    kind: ClassVar[ColorMapKind] = ColorMapKind.miscellaneous

    def _generate_colors(self, x: FloatArrayLike) -> Sequence[RGBHexColor]:
        x = np.asarray(x)
        # Apply gamma factor to emphasise low or high intensity values
        xg = x**self.gamma

        # Calculate amplitude and angle of deviation from the black to
        # white diagonal in the plane of constant perceived intensity.
        amplitude = self.hue * xg * (1 - xg) / 2
        phi = 2 * np.pi * (self.start / 3 + self.rotation * x)

        sin_cos = np.array([np.cos(phi), np.sin(phi)])
        rgb = (xg + amplitude * np.dot(rotation_matrix, sin_cos)).T

        if self.reverse:
            rgb = rgb[::-1, :]

        return [rgb_to_hex(c) for c in rgb]

    def discrete_palette(self, n: int) -> Sequence[RGBHexColor]:
        """
        Return n colors from the gradient

        Parameters
        ----------
        n :
            Number of colors to return from the gradient.
        """
        x = np.linspace(self.light, self.dark, n)
        return self._generate_colors(x)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from ._colormap import ColorMapKind
from ._interpolated import _InterpolatedGen, interp_lookup

if TYPE_CHECKING:
    from mizani.typing import SegmentedColorMapData


@dataclass
class SegmentInterpolatedMap(_InterpolatedGen):
    """
    Gradient colormap by interpolating RGB colors independently

    The input data is the same as Matplotlib's LinearSegmentedColormap
    data.
    """

    data: SegmentedColorMapData
    kind: ColorMapKind = ColorMapKind.miscellaneous

    def __post_init__(self):
        _red = np.asarray(self.data["red"])
        _blue = np.asarray(self.data["blue"])
        _green = np.asarray(self.data["green"])
        self._r_lookup = interp_lookup(*_red.T)
        self._g_lookup = interp_lookup(*_green.T)
        self._b_lookup = interp_lookup(*_blue.T)
        self.colors = self.data



"""
According to *On the theory of scales of measurement* by **S.S. Stevens**,
scales can be classified in four ways -- *nominal*, *ordinal*,
*interval* and *ratio*. Using current(2016) terminology, *nominal* data
is made up of unordered categories, *ordinal* data is made up of ordered
categories and the two can be classified as *discrete*. On the other hand
both *interval* and *ratio* data are *continuous*.

The scale classes below show how the rest of the Mizani package can be
used to implement the two categories of scales. The key tasks are
*training* and *mapping* and these correspond to the **train** and
**map** methods.

To train a scale on data means, to make the scale learn the limits of
the data. This is elaborate (or worthy of a dedicated method) for two
reasons:

    - *Practical* -- data may be split up across more than one object,
      yet all will be represented by a single scale.
    - *Conceptual* -- training is a key action that may need to be inserted
      into multiple locations of the data processing pipeline before a
      graphic can be created.

To map data onto a scale means, to associate data values with
values(potential readings) on a scale. This is perhaps the most important
concept unpinning a scale.

The **apply** methods are simple examples of how to put it all together.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
import pandas as pd

from .bounds import censor, rescale
from .utils import (
    CONTINUOUS_KINDS,
    DISCRETE_KINDS,
    get_categories,
    has_dtype,
    match,
    min_max,
)

if TYPE_CHECKING:
    from typing import Any, Optional, Sequence, TypeVar

    from mizani.typing import (
        AnyArrayLike,
        Callable,
        ContinuousPalette,
        DiscretePalette,
        FloatArrayLike,
        NDArrayFloat,
        Trans,
        TupleFloat2,
    )

    TVector = TypeVar("TVector", NDArrayFloat, pd.Series[float])


__all__ = ["scale_continuous", "scale_discrete"]


class scale_continuous:
    """
    Continuous scale
    """

    @classmethod
    def apply(
        cls,
        x: FloatArrayLike,
        palette: ContinuousPalette,
        na_value: Any = None,
        trans: Optional[Trans] = None,
    ) -> NDArrayFloat:
        """
        Scale data continuously

        Parameters
        ----------
        x : array_like
            Continuous values to scale
        palette : callable ``f(x)``
            Palette to use
        na_value : object
            Value to use for missing values.
        trans : trans
            How to transform the data before scaling. If
            ``None``, no transformation is done.

        Returns
        -------
        out : array_like
            Scaled values
        """
        if trans is not None:
            x = trans.transform(x)

        limits = cls.train(x)
        return cls.map(x, palette, limits, na_value)

    @classmethod
    def train(
        cls, new_data: FloatArrayLike, old: Optional[TupleFloat2] = None
    ) -> TupleFloat2:
        """
        Train a continuous scale

        Parameters
        ----------
        new_data : array_like
            New values
        old : array_like
            Old range

        Returns
        -------
        out : tuple
            Limits(range) of the scale
        """
        if old is None:
            old = (-np.inf, np.inf)

        if not len(new_data):
            return old

        new_data = np.asarray(new_data)

        if new_data.dtype.kind not in CONTINUOUS_KINDS:
            raise TypeError("Discrete value supplied to continuous scale")

        new_data = np.hstack([new_data, old])
        return min_max(new_data, na_rm=True, finite=True)

    @classmethod
    def map(
        cls,
        x: FloatArrayLike,
        palette: ContinuousPalette,
        limits: TupleFloat2,
        na_value: Any = None,
        oob: Callable[[TVector], TVector] = censor,
    ) -> NDArrayFloat:
        """
        Map values to a continuous palette

        Parameters
        ----------
        x : array_like
            Continuous values to scale
        palette : callable ``f(x)``
            palette to use
        na_value : object
            Value to use for missing values.
        oob : callable ``f(x)``
            Function to deal with values that are
            beyond the limits

        Returns
        -------
        out : array_like
            Values mapped onto a palette
        """
        x = oob(rescale(x, _from=limits))  # pyright: ignore
        pal = np.asarray(palette(x))
        pal[pd.isna(x)] = na_value
        return pal


class scale_discrete:
    """
    Discrete scale
    """

    @classmethod
    def apply(
        cls,
        x: AnyArrayLike,
        palette: DiscretePalette,
        na_value: Any = None,
    ):
        """
        Scale data discretely

        Parameters
        ----------
        x : array_like
            Discrete values to scale
        palette : callable ``f(x)``
            Palette to use
        na_value : object
            Value to use for missing values.

        Returns
        -------
        out : array_like
            Scaled values
        """
        limits = cls.train(x)
        return cls.map(x, palette, limits, na_value)

    @classmethod
    def train(
        cls,
        new_data: AnyArrayLike,
        old: Optional[Sequence[Any]] = None,
        drop: bool = False,
        na_rm: bool = False,
    ) -> Sequence[Any]:
        """
        Train a continuous scale

        Parameters
        ----------
        new_data : array_like
            New values
        old : array_like
            Old range. List of values known to the scale.
        drop : bool
            Whether to drop(not include) unused categories
        na_rm : bool
            If ``True``, remove missing values. Missing values
            are either ``NaN`` or ``None``.

        Returns
        -------
        out : list
            Values covered by the scale
        """
        old = [] if old is None else list(old)

        if not len(new_data):
            return old

        old_set = set(old)

        # Get the missing values (NaN & Nones) locations and remove them
        nan_bool_idx = pd.isna(new_data)  # type: ignore
        has_na = np.any(nan_bool_idx)

        if not has_dtype(new_data):
            new_data = np.asarray(new_data)

        new_data = cast(np.ndarray, new_data)

        if new_data.dtype.kind not in DISCRETE_KINDS:
            raise TypeError("Continuous value supplied to discrete scale")

        new_data = new_data[~nan_bool_idx]

        # 1. Train i.e. get the new values
        # 2. Update old
        if isinstance(new_data.dtype, pd.CategoricalDtype):
            categories = get_categories(new_data)
            if drop:
                present = set(new_data)
                new = [i for i in categories if i in present]
            else:
                new = list(categories)

            all_set = old_set | set(new)
            ordered_cats = categories.union(old, sort=False)
            limits = [c for c in ordered_cats if c in all_set]
        else:
            new = np.unique(new_data)
            new.sort()

            limits = old + [i for i in new if (i not in old_set)]

        # Add nan if required
        has_na_limits = any(pd.isna(limits))
        if not has_na_limits and not na_rm and has_na:
            limits.append(np.nan)
        return limits

    @classmethod
    def map(
        cls,
        x: AnyArrayLike,
        palette: DiscretePalette,
        limits: Sequence[Any],
        na_value: Any = None,
    ) -> AnyArrayLike:
        """
        Map values to a discrete palette

        Parameters
        ----------
        palette : callable ``f(x)``
            palette to use
        x : array_like
            Continuous values to scale
        na_value : object
            Value to use for missing values.

        Returns
        -------
        out : array_like
            Values mapped onto a palette
        """
        n = len(limits)
        pal = np.asarray(palette(n))[match(x, limits)]
        nas = pd.isna(x)  # type: ignore
        try:
            pal[nas] = na_value
        except TypeError:
            pal = [na_value if isna else v for v, isna in zip(pal, nas)]

        return pal



from mizani._colors import SegmentFunctionMap
from mizani._colors._colormaps._maps._segment_function import GPF, _gpf_32


def test_discrete():
    cmap = SegmentFunctionMap(
        {"red": _gpf_32, "green": GPF[3], "blue": GPF[17]}
    )
    colors = cmap.discrete_palette(5)
    assert len(colors) == 5
    assert [c[0] == "#" and len(c) == 7 for c in colors]



from mizani._colors._palettes.brewer import (
    diverging,
    qualitative,
    sequential,
)
from mizani._colors._palettes.other import (
    qualitative as other_qualitative,
)

__all__ = (
    # diverging
    "BrBG",
    "PiYG",
    "PRGn",
    "PuOr",
    "RdBu",
    "RdGy",
    "RdYlBu",
    "RdYlGn",
    "Spectral",
    # qualitative
    "Accent",
    "Dark2",
    "Paired",
    "Pastel1",
    "Pastel2",
    "Set1",
    "Set2",
    "Set3",
    # sequential
    "Blues",
    "BuGn",
    "BuPu",
    "GnBu",
    "Greens",
    "Greys",
    "Oranges",
    "OrRd",
    "PuBu",
    "PuBuGn",
    "PuRd",
    "Purples",
    "RdPu",
    "Reds",
    "YlGn",
    "YlGnBu",
    "YlOrBr",
    "YlOrRd",
    # qualitative
    "category10",
    "category20",
    "category20b",
    "category20c",
    "observable10",
    "tableau10",
    "tableau20",
)

BrBG = diverging.BrBG.colormap
PiYG = diverging.PiYG.colormap
PRGn = diverging.PRGn.colormap
PuOr = diverging.PuOr.colormap
RdBu = diverging.RdBu.colormap
RdGy = diverging.RdGy.colormap
RdYlBu = diverging.RdYlBu.colormap
RdYlGn = diverging.RdYlGn.colormap
Spectral = diverging.Spectral.colormap

Accent = qualitative.Accent.colormap
Dark2 = qualitative.Dark2.colormap
Paired = qualitative.Paired.colormap
Pastel1 = qualitative.Pastel1.colormap
Pastel2 = qualitative.Pastel2.colormap
Set1 = qualitative.Set1.colormap
Set2 = qualitative.Set2.colormap
Set3 = qualitative.Set3.colormap

Blues = sequential.Blues.colormap
BuGn = sequential.BuGn.colormap
BuPu = sequential.BuPu.colormap
GnBu = sequential.GnBu.colormap
Greens = sequential.Greens.colormap
Greys = sequential.Greys.colormap
Oranges = sequential.Oranges.colormap
OrRd = sequential.OrRd.colormap
PuBu = sequential.PuBu.colormap
PuBuGn = sequential.PuBuGn.colormap
PuRd = sequential.PuRd.colormap
Purples = sequential.Purples.colormap
RdPu = sequential.RdPu.colormap
Reds = sequential.Reds.colormap
YlGn = sequential.YlGn.colormap
YlGnBu = sequential.YlGnBu.colormap
YlOrBr = sequential.YlOrBr.colormap
YlOrRd = sequential.YlOrRd.colormap

category10 = other_qualitative.category10.colormap
category20 = other_qualitative.category20.colormap
category20b = other_qualitative.category20b.colormap
category20c = other_qualitative.category20c.colormap
observable10 = other_qualitative.observable10.colormap
tableau10 = other_qualitative.tableau10.colormap
tableau20 = other_qualitative.tableau20.colormap



from mizani._colors import CubeHelixMap


def test_continuous():
    x = [0.1, 0.2, 0.3, 0.4, 0.5]
    chmap = CubeHelixMap(reverse=True)
    colors = chmap.continuous_palette(x)
    assert len(colors) == 5
    assert all(c is not None and c.startswith("#") for c in colors)

    x = [0.1, 0.2, float("nan"), 0.4, float("inf")]
    colors = chmap.continuous_palette(x)
    assert len(colors) == 5
    assert colors[2] is None
    assert colors[4] is None
    assert all(colors[i].startswith("#") for i in (0, 1, 3))



from __future__ import annotations

import sys
from enum import Enum
from typing import Optional

RESET = "\033[0m"


class Fg(Enum):
    """
    Foreground color codes
    """

    black = "\033[30m"
    red = "\033[31m"
    green = "\033[32m"
    orange = "\033[33m"
    blue = "\033[34m"
    purple = "\033[35m"
    cyan = "\033[36m"
    lightgrey = "\033[37m"
    darkgrey = "\033[90m"
    lightred = "\033[91m"
    lightgreen = "\033[92m"
    yellow = "\033[93m"
    lightblue = "\033[94m"
    pink = "\033[95m"
    lightcyan = "\033[96m"


class Bg(Enum):
    """
    Background color codes
    """

    black = "\033[40m"
    red = "\033[41m"
    green = "\033[42m"
    orange = "\033[43m"
    blue = "\033[44m"
    purple = "\033[45m"
    cyan = "\033[46m"
    lightgrey = "\033[47m"


class Effect(Enum):
    """
    Text effect codes
    """

    bold = "\033[01m"
    dim = "\033[02m"
    underline = "\033[04m"
    blink = "\033[05m"
    reverse = "\033[07m"  # bg & fg are reversed
    hide = "\033[08m"
    strikethrough = "\033[09m"


def T(
    s: str,
    fg: Optional[str] = None,
    bg: Optional[str] = None,
    effect: Optional[str] = None,
) -> str:
    """
    Enclose text string with ANSI codes

    e.g.
        # Red text
        T("sample", "red")

        # Red on lightgrey background
        T("sample", "red", "lightgrey")

        # Red on lightgrey background and underlined
        T("sample", "red", "lightgrey", "underlined")

        # Red underlined text
        T("sample", effect="underlined")

        # Red & bold underlined text
        T("sample", effect="bold, underlined")
    """

    def get(Ecls, prop_name) -> str:
        return getattr(Ecls, prop_name).value if prop_name else ""

    _fg = get(Fg, fg)
    _bg = get(Bg, bg)
    if effect:
        _effect = "".join(get(Effect, e.strip()) for e in effect.split(","))
    else:
        _effect = ""

    _reset = RESET if any((_fg, _bg, _effect)) else ""
    return f"{_fg}{_bg}{_effect}{s}{_reset}"


def T0(s: str, *args, **kwargs) -> str:
    """
    Enclose text string with ANSI codes output is tty
    """
    if sys.stdout.isatty():
        return T(s, *args, **kwargs)
    return s



import warnings
from datetime import datetime, timedelta

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from numpy.testing import assert_equal
from pytest import approx

from mizani.bounds import (
    censor,
    expand_range,
    expand_range_distinct,
    rescale,
    rescale_max,
    rescale_mid,
    squish,
    squish_infinite,
    zero_range,
)

NaT_type = type(pd.NaT)


def test_censor():
    x = np.arange(10)
    xx = censor(x, (2, 8))
    assert np.isnan(xx[0])
    assert np.isnan(xx[1])
    assert np.isnan(xx[9])

    data = pd.DataFrame({"x": x, "y": range(10)})
    data["x"] = censor(data["x"], (2, 8))
    assert np.isnan(data["x"][0])
    assert np.isnan(data["x"][1])
    assert np.isnan(data["x"][9])

    data["y"] = censor(data["y"], (-2, 18))
    assert issubclass(data["y"].dtype.type, np.integer)

    # datetime
    limits = datetime(2010, 1, 1), datetime(2020, 1, 1)
    x = np.array([datetime(year, 1, 1) for year in range(2008, 2023)])
    result = censor(x, limits)
    assert_equal(result[2:-2], x[2:-2])
    assert_equal(result[:2], [None, None])
    assert_equal(result[-2:], [None, None])

    # timedelta
    limits = timedelta(seconds=2010), timedelta(seconds=2020)
    x = np.array([timedelta(seconds=i) for i in range(2008, 2023)])
    result = censor(x, limits)
    assert_equal(result[2:-2], x[2:-2])
    assert_equal(result[:2], [None, None])
    assert_equal(result[-2:], [None, None])

    # pd.timestamp
    limits = pd.Timestamp(200 * 1e16), pd.Timestamp(205 * 1e16)
    x = np.array([pd.Timestamp(i * 1e16) for i in range(198, 208)])
    result = censor(x, limits)
    assert_equal(result[2:-2], x[2:-2])
    assert_equal(result[:2], [None, None])
    assert_equal(result[-2:], [None, None])

    x1 = np.array(x)
    result = censor(x1, limits)
    assert all(val is None for val in result[:2])
    assert all(val is None for val in result[-2:])

    x2 = pd.Series(x)
    result = censor(x2, limits)
    pdt.assert_series_equal(result[2:-2], x2[2:-2])
    assert all(isinstance(val, NaT_type) for val in result[:2])
    assert all(isinstance(val, NaT_type) for val in result[-2:])

    # np.datetime
    limits = np.datetime64(200, "D"), np.datetime64(205, "D")
    x = [np.datetime64(i, "D") for i in range(198, 208)]
    x2 = np.array(x)
    result = censor(x2, limits)
    assert_equal(result[2:-2], x2[2:-2])
    assert all(isinstance(val, np.datetime64) for val in result[:2])
    assert all(isinstance(val, np.datetime64) for val in result[-2:])

    # pd.Timedelta
    limits = pd.Timedelta(seconds=2010), pd.Timedelta(seconds=2020)
    x = np.array([pd.Timedelta(seconds=i) for i in range(2008, 2023)])
    result = censor(x, limits)
    assert_equal(result[2:-2], x[2:-2])
    assert all(val is None for val in result[:2])
    assert all(val is None for val in result[-2:])

    x4 = np.array(x)
    result = censor(x4, limits)
    npt.assert_array_equal(result[2:-2], x4[2:-2])
    assert all(val is None for val in result[:2])
    assert all(val is None for val in result[-2:])

    # np.timedelta64
    limits = np.timedelta64(200, "D"), np.timedelta64(205, "D")
    x = [np.timedelta64(i, "D") for i in range(198, 208)]
    x5 = np.array(x)
    result = censor(x5, limits)
    npt.assert_array_equal(result[2:-2], x5[2:-2])
    assert all(isinstance(val, np.timedelta64) for val in result[:2])
    assert all(isinstance(val, np.timedelta64) for val in result[-2:])

    # branches #
    x = np.array([1, 2, np.inf, 3, 4, 11])
    result = censor(x, (0, 10), only_finite=False)
    npt.assert_array_equal(result, np.array([1, 2, np.nan, 3, 4, np.nan]))

    result = censor([], (-2, 18))
    assert len(result) == 0

    with pytest.raises(ValueError):
        result = censor(["a", "b", "c"], ("a", "z"))


def test_expand_range():
    assert expand_range((0, 1)) == (0, 1)
    assert expand_range((0, 1), mul=2) == (-2, 3)
    assert expand_range((0, 1), add=2) == (-2, 3)
    assert expand_range((0, 1), mul=2, add=2) == (-4, 5)
    assert expand_range((1, 1), mul=2, add=2, zero_width=1) == (0.5, 1.5)
    assert expand_range((0, 0)) == (-0.5, 0.5)

    def diff(x):
        return x[1] - x[0]

    # datetime
    one_day = datetime(2010, 1, 2) - datetime(2010, 1, 1)
    limits = datetime(2010, 1, 1), datetime(2010, 1, 2)
    result = expand_range(limits, add=one_day)
    diff(result) == diff(limits) + 2 * one_day

    limits = datetime(2010, 1, 1), datetime(2010, 1, 1)  # zero range
    result = expand_range(limits, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # pd.Timestamp
    one_day = pd.Timestamp("2010-01-02") - pd.Timestamp("2010-01-01")
    limits = pd.Timestamp("2010-01-01"), pd.Timestamp("2010-01-02")
    result = expand_range(limits, add=one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = pd.Timestamp("2010-01-01"), pd.Timestamp("2010-01-01")
    result = expand_range(limits, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # np.datetime64
    one_day = np.datetime64(1, "D") - np.datetime64(0, "D")
    limits = np.datetime64(14610, "D"), np.datetime64(14611, "D")
    result = expand_range(limits, add=one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = np.datetime64(14610, "D"), np.datetime64(14611, "D")
    result = expand_range(limits, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # timedelta
    one_day = timedelta(days=1)
    limits = timedelta(days=1), timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = timedelta(days=10), timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # pd.Timedelta
    one_day = pd.Timedelta(days=1)
    limits = pd.Timedelta(days=1), pd.Timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = pd.Timedelta(days=10), pd.Timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # timedelta64
    one_day = np.timedelta64(1, "D")
    limits = np.timedelta64(1, "D"), np.timedelta64(10, "D")
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = np.timedelta64(1, "D"), np.timedelta64(1, "D")
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day


def test_expand_range_distinct():
    assert expand_range_distinct((0, 1)) == (0, 1)
    assert expand_range_distinct((0, 1), (2, 0)) == (-2, 3)
    assert expand_range_distinct((0, 1), (2, 0, 2, 0)) == (-2, 3)
    assert expand_range_distinct((0, 1), (0, 2)) == (-2, 3)
    assert expand_range_distinct((0, 1), (0, 2, 0, 2)) == (-2, 3)
    assert expand_range_distinct((0, 1), (2, 2, 2, 2)) == (-4, 5)
    assert expand_range_distinct((1, 1), (2, 2), zero_width=1) == (0.5, 1.5)


def test_rescale():
    # [0, 10] -> [0, 1]
    # Results are invariant to uniformly translated
    # or expanded inputs
    a = np.arange(0, 11)
    npt.assert_allclose(rescale(a), a * 0.1)
    npt.assert_allclose(rescale(a), rescale(a - 42))
    npt.assert_allclose(rescale(a), rescale(a + 42))
    npt.assert_allclose(rescale(a), rescale(a / np.pi))
    npt.assert_allclose(rescale(a), rescale(a * np.pi))

    # Some more
    n = 6
    a = np.arange(0, n)
    npt.assert_allclose(rescale(a), a / (n - 1))
    npt.assert_allclose(rescale(a, _from=(0, 10)), a * 0.1)
    npt.assert_allclose(rescale(a, to=(0, n * (n - 1))), a * n)


def test_rescale_max():
    a = np.arange(0, 11)
    # Max & first element has no effect
    assert max(rescale_max(a, to=(0, 5))) == 5
    assert max(rescale_max(a, to=(4, 5))) == 5

    # Expanded & first elements have no effect
    assert max(rescale_max(a, to=(0, 5), _from=(0, 3))) > 5
    assert max(rescale_max(a, to=(2, 5), _from=(2, 3))) > 5

    # branches #
    assert rescale_max(2, _from=(0, 10)) == 0.2

    x = [-5, -4, -3, -2, -1, 0]
    result = rescale_max(x)
    assert result == approx([0, 0.2, 0.4, 0.6, 0.8, 1.0])

    result = rescale_max(x, to=(0, 5))
    assert result == approx([0, 1, 2, 3, 4, 5])


def test_rescale_mid():
    a = [1, 2, 3]
    # no change
    npt.assert_allclose(rescale_mid(a, to=(1, 3), mid=2), a)
    npt.assert_allclose(rescale_mid(a, to=(1, 4), _from=(1, 4), mid=2.5), a)

    npt.assert_allclose(rescale_mid(a, mid=1), [0.5, 0.75, 1])
    npt.assert_allclose(rescale_mid(a, mid=2), [0, 0.5, 1])
    npt.assert_allclose(rescale_mid(a, mid=3), [0, 0.25, 0.5])

    # branches #
    npt.assert_allclose(rescale_mid([2], _from=(2, 2), to=(2, 2), mid=2), [2])


def test_squish_infinite():
    a = [-np.inf, np.inf, -np.inf, np.inf]
    npt.assert_allclose(squish_infinite(a), [0, 1, 0, 1])
    npt.assert_allclose(
        squish_infinite(a, (-100, 100)), [-100, 100, -100, 100]
    )

    b = np.array([5, -np.inf, 2, 3, 6])
    npt.assert_allclose(squish_infinite(b, (1, 10)), [5, 1, 2, 3, 6])


def test_squish():
    a = [-np.inf, np.inf, -np.inf, np.inf]
    npt.assert_allclose(squish(a, only_finite=False), [0, 1, 0, 1])
    npt.assert_allclose(squish(a, only_finite=True), a)
    npt.assert_allclose(
        squish(a, (-100, 100), only_finite=False), [-100, 100, -100, 100]
    )

    b = np.array([5, 0, -2, 3, 10])
    npt.assert_allclose(squish(b, (0, 5)), [5, 0, 0, 3, 5])

    c = np.array([5, -np.inf, 2, 3, 6])
    npt.assert_allclose(squish(c, (1, 10), only_finite=False), [5, 1, 2, 3, 6])
    npt.assert_allclose(squish(c, (1, 10)), c)


def test_zero_range():
    c = np.array
    eps = np.finfo(float).eps

    assert zero_range(c((1, 1 + eps)))
    assert zero_range(c((1, 1 + 99 * eps)))

    # Crossed the tol threshold
    # assert zero_range(c((1, 1 + 101 * eps))) is False
    assert not zero_range(c((1, 1 + 101 * eps)))

    # Changed tol
    assert not zero_range(c((1, 1 + 2 * eps)), tol=eps)

    # Scaling up or down all the values has no effect since
    # the values are rescaled to 1 before checking against
    # the tolerance
    assert zero_range(100000 * c((1, 1 + eps)))
    assert not zero_range(100000 * c((1, 1 + 200 * eps)))
    assert zero_range(0.00001 * c((1, 1 + eps)))
    assert not zero_range(0.00001 * c((1, 1 + 200 * eps)))

    # NA values
    assert zero_range((1, np.nan))

    # Infinite values
    assert not zero_range((1, np.inf))
    assert not zero_range((-np.inf, np.inf))
    assert zero_range((np.inf, np.inf))

    # Single value
    with pytest.raises(TypeError):
        assert zero_range(1)

    with pytest.raises(IndexError):
        assert zero_range([1])

    # length greater than 2
    with pytest.raises(ValueError):
        zero_range([1, 2, 3])

    # datetime - pandas, cpython
    x = datetime(2010, 1, 1), datetime(2010, 1, 1)
    x2 = datetime(2010, 1, 1), datetime(2020, 1, 1)
    x3 = (
        pd.Timestamp("2010-01-01", tz="US/Eastern"),
        pd.Timestamp("2010-01-01", tz="US/Central"),
    )
    assert zero_range(x)
    assert not zero_range(x2)
    assert not zero_range(x3)

    # datetime - numpy
    x = np.datetime64(7, "D"), np.datetime64(7, "D")
    x2 = np.datetime64(7, "D"), np.datetime64(1, "W")
    x3 = np.datetime64(7, "D"), np.datetime64(1, "D")
    assert zero_range(x)
    assert zero_range(x2)
    assert not zero_range(x3)

    # timedelta - pandas, cpython
    x = timedelta(seconds=2010), timedelta(seconds=2010)
    x2 = (
        timedelta(seconds=2010, microseconds=90),
        timedelta(seconds=2010, microseconds=34),
    )
    x3 = pd.Timedelta(200, "D"), pd.Timedelta(203, "D")
    assert zero_range(x)
    assert not zero_range(x2)
    assert not zero_range(x3)

    # timedelta - numpy
    x = np.timedelta64(7, "D"), np.timedelta64(7, "D")
    x2 = np.timedelta64(7, "D"), np.timedelta64(1, "W")
    x3 = np.timedelta64(7, "D"), np.timedelta64(2, "D")
    assert zero_range(x)
    assert zero_range(x2)
    assert not zero_range(x3)

    # branches #
    assert zero_range([4, float("nan")])
    assert not zero_range([4, float("inf")])
    with pytest.raises(TypeError):
        zero_range(["a", "b"])

    # are handled without error
    x = np.array([8, 4], dtype=np.uint64)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        zero_range(x)
